from musikla.parser.printer import CodePrinter
from typing import Tuple
from .expression_node import ExpressionNode
from musikla.core import Value

class NoneLiteralNode( ExpressionNode ):
    def __init__ ( self, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.value = None

    def __eval__ ( self, context, assignment : bool = False ):
        return self.value
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( "none" )



from typing import Any, Optional, Tuple, Union, cast
from musikla.core import Context, Music
from musikla.core.events import MusicEvent
from musikla.core.events.transformers import SortTransformer
from .keyboard import Keyboard
from fractions import Fraction
import asyncio

DIRECTION_LEFT = 'left'
DIRECTION_RIGHT = 'right'
DIRECTION_BOTH = 'both'

class Grid:
    def __init__ ( self, context : Context, num : int = 1, den : int = 1, forgiveness : int = 0, forgiveness_left : int = 0, forgiveness_right : int = 0, range : int = None, range_left : int = None, range_right : int = None, direction : str = DIRECTION_BOTH, sync_with : 'Grid' = None, prealign_with : 'Grid' = None ):
        self.context : Context = context
        self.length : Fraction = Fraction( num, den )
        self.length_duration : int = self.context.voice.get_duration( float( self.length ) )
        
        self._start : Optional[int] = None
        self.realtime : bool = False

        self.range : Optional[int] = range
        self.range_left : Optional[int] = range_left
        self.range_right : Optional[int] = range_right

        # A natural number specifying the nonaligment forgiveness: if the event is badly aligned with the grid but the
        # distance is less than the forgiveness set, in milliseconds, then no alignment is performed.
        self.forgiveness : int = forgiveness
        self.forgiveness_left : int = forgiveness_left
        self.forgiveness_right : int = forgiveness_right

        # Determines, when aligning a music event, what direction the alignment can go.
        # Left means that an event is always realigned to the nearest grid elemente befor it
        # Right means that an event is always realigned to the nearest grid element after it
        # Both means that an event is always realigned to the grid element that is closest to it
        self.direction : str = direction

        self.sync_with : Optional['Grid'] = sync_with
        self.prealign_with : Optional['Grid'] = prealign_with

    @property
    def player ( self ):
        from .library import KeyboardLibrary

        library = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

        return library.player

    @property
    def start ( self ) -> Optional[int]:
        if self.sync_with is not None:
            return self.sync_with.start

        return self._start

    def _real_ranges ( self ) -> Tuple[int, int]:
        range = int( self.length_duration / 2 ) if self.range is None else self.range
        
        range_left = range if self.range_left is None else self.range_left
        
        range_right = range if self.range_right is None else self.range_right
        
        if self.range_left is None and self.range_right is not None:
            range_left = range * 2 - self.range_right
        elif self.range_left is not None and self.range_right is None:
            range_right = range * 2 - self.range_left

        return ( range_left, range_right )

    def _real_forgiveness ( self ) -> Tuple[int, int]:
        forgiveness = 0 if self.forgiveness is None else self.forgiveness

        forgiveness_left = forgiveness if self.forgiveness_left is None else self.forgiveness_left
        
        forgiveness_right = forgiveness if self.forgiveness_right is None else self.forgiveness_right
        
        return ( forgiveness_left, forgiveness_right )

    def _within_forgiveness ( self, dist_l : int, dist_r : int ) -> bool:
        fl, fr = self._real_forgiveness()

        return dist_r <= fl or abs( dist_l ) <= fr

    def _within_range ( self, dist_l : int, dist_r : int ) -> int:
        rl, rr = self._real_ranges()

        if dist_l <= rr:
            return dist_r if self.direction == 'right' else -dist_l
        elif dist_r <= rl:
            return -dist_l if self.direction == 'left' else dist_r
        else:
            return 0
    
    def get_delta ( self, time : int ) -> int:
        rest = ( time - self.start ) % self.length_duration

        if rest == 0:
            return 0

        dist_l = rest
        dist_r = self.length_duration - rest
            
        if self._within_forgiveness( dist_l, dist_r ):
            return 0

        return self._within_range( dist_l, dist_r )

    def reset ( self, base : int = None ):
        if self.sync_with is not None:
            self.sync_with.reset( base )
        else:
            self._start = base

    def _align_event ( self, event : MusicEvent, start_time : int, individually : bool = False, maintain_duration : bool = True ) -> MusicEvent:
        start = self.start

        if start is None:
            self.reset( start_time )

        if individually:
            event_start_delta = self.get_delta( event.timestamp )
        else:
            event_start_delta = self.get_delta( start_time )

        event_end_delta = self.get_delta( event.end_timestamp + event_start_delta ) if not maintain_duration and individually else 0

        return event.clone( 
            timestamp = event_start_delta + event.timestamp, 
            end_timestamp = event_start_delta + event_end_delta + event.end_timestamp 
        )

    def align ( self, context : Context, music : Union[Music, MusicEvent, Any], mode : str = 'start_end', individually : bool = False, maintain_duration : bool = True ) -> Union[Music, MusicEvent, Any]:
        """
        Allowed modes are:
         - global: Changes only the sequence's start time
         - start: Changes the note's start time but retains their duration
         - start_end: Changes both the note's start time and duration to be aligned with the grid
        """

        if self.prealign_with is not None:
            music = self.prealign_with.align( context, music, mode )

        individually = mode != 'global'
        maintain_duration = not individually or mode != 'start_end'

        if isinstance( music, Music ):
            aligned : Music = music.map( lambda e, i, s: self._align_event( e, s, individually, maintain_duration ).join( context ) )

            if individually:
                forgiveness = self.forgiveness + self.forgiveness_left + self.forgiveness_right

                if forgiveness != 0:
                    aligned = aligned.transform( SortTransformer, forgiveness )

            return aligned
        elif isinstance( music, MusicEvent ):
            return self._align_event( music, music.timestamp, individually, maintain_duration )
        else:
            return music

        # if fastforward:
            # aligned = shared = aligned.shared()

            # first_event = shared.peek()

            # if first_event is not None:
            # now = self.player.get_time() - self.start

            # aligned = aligned.slice( start = now, time = True, cut = True )

    @property
    def cell_time ( self ) -> int:
        if self.start is None:
            return 0

        return ( self.player.get_time() - self.start ) % self.length_duration

    def cli_metronome ( self, toolbar_width = 100 ):
        async def periodic():
            while True:
                toolbar_filled = int( self.cell_time * toolbar_width / self.length_duration )
                toolbar_remaining = toolbar_width - toolbar_filled

                print( f'\r[{ "-" * toolbar_filled }{ " " * toolbar_remaining }]    ', end = '' )

                await asyncio.sleep(0.01)

        asyncio.create_task( periodic() )

    @staticmethod
    def compose ( *args ):
        last_grid = args[ 0 ]

        for i in range( 1, len( args ) ):
            args[ i ].sync_with = last_grid.sync_with if last_grid.sync_with is not None else last_grid
            args[ i ].prealign_with = last_grid

            last_grid = args[ i ]

        return last_grid


from musikla.parser.printer import CodePrinter
from musikla.parser.abstract_syntax_tree.expressions.variable_expression_node import VariableExpressionNode
from musikla.parser.abstract_syntax_tree.expressions.property_accessor_node import PropertyAccessorNode
from .expression_node import ExpressionNode
from musikla.core import Value
from musikla.core.callable_python_value import CallablePythonValue
from typing import Callable, Tuple, cast

class FunctionChainExpressionNode( ExpressionNode ):
    def __init__ ( self, left_hand, right_hand, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.left_hand = left_hand
        self.right_hand = right_hand

        self.right_hand.expression = left_hand
        
        if position is not None:
            self.right_hand.position = position
    
    @property
    def position ( self ):
        return self.right_hand.position

    @position.setter
    def position( self, value ):
        if hasattr( self, 'right_hand' ) and self.right_hand is not None:
            self.right_hand.position = value

    @property
    def expression ( self ):
        return self.left_hand.expression

    @expression.setter
    def expression( self, value ):
        self.left_hand.expression = value

    def __eval__ ( self, context ):
        return self.right_hand.eval( context )

    def to_source ( self, printer : CodePrinter ):
        self.right_hand.to_source( printer )

class FunctionExpressionNode( ExpressionNode ):
    def __init__ ( self, expression, parameters = [], named_parameters = dict(), position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.expression : ExpressionNode = expression
        self.parameters = parameters
        self.named_parameters = named_parameters

    def __eval__ ( self, context ):
        value = self.expression.eval( context )
        
        if value == None: 
            if isinstance( self.expression, VariableExpressionNode ):
                raise BaseException( f"Calling undefined function { cast( VariableExpressionNode, self.expression ).name }" )
            elif isinstance( self.expression, PropertyAccessorNode ):
                method_name = cast( PropertyAccessorNode, self.expression ).name.eval( context )
                object_name = CodePrinter().print( cast( PropertyAccessorNode, self.expression ).expression )

                raise BaseException( f"Calling undefined function '{ method_name }' in '{ object_name }'" )
            else:
                raise BaseException( "Calling undefined function" )

        return CallablePythonValue.call( value, context, self.parameters, self.named_parameters )

    def to_source ( self, printer : CodePrinter ):
        from .variable_expression_node import VariableExpressionNode

        if isinstance( self.expression, VariableExpressionNode ):
            printer.add_token( self.expression.name )
        else:
            self.expression.to_source( printer )
        
        with printer.block( '(', ')' ):
            for i in range( len( self.parameters ) ):
                if i > 0:
                    printer.add_token( '; ' )
                
                self.parameters[ i ].to_source( printer )

            not_first = bool( self.parameters )

            for key, node in self.named_parameters.items():
                if not_first:
                    printer.add_token( '; ' )

                printer.add_token( key + ' = ' )
                
                node.to_source( printer )

                not_first = True



from musikla.parser.printer import CodePrinter
from typing import Tuple, Union, Any
from .node import Node
from musikla.core import Context
import inspect

class PythonNode( Node ):
    
    @staticmethod
    def get_auto_name ( val ):
        return val.__name__

    @staticmethod
    def create_export_decorator ( context : Context ):
        def export ( name : Union[str, any] = None ):
            nonlocal context

            def export_instance ( val ):
                nonlocal name, context

                if name is None:
                    name = PythonNode.get_auto_name( val )
                
                context.symbols.assign( name, val )

                return val
            
            if name is not None and type(name) is not str:
                val, name = name, None

                return export_instance( val )
            else:
                return export_instance

        return export

    def execute ( context : Context, code : Union[str, Any], is_expression : bool = False ):
        if type( code ) is str:
            code = compile( code, "<embedded python>", 'eval' if is_expression else 'exec' )

        custom_globals = {}

        if is_expression:
            custom_locals = PythonContext( context )
        else:
            custom_globals[ 'export' ] = PythonNode.create_export_decorator( context )
            custom_locals = {}

        return eval( code, custom_globals, custom_locals )

    def __init__ ( self, code : str, is_expression : bool = False, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.code : str = code
        self.is_expression : bool = is_expression
        self.bytecode = compile( self.code, "<embedded python>", 'eval' if is_expression else 'exec' )
        self.hoisted : bool = not is_expression
    
    def to_source ( self, printer : CodePrinter ):
        if self.is_expression:
            printer.add_token( '@py' )
            with printer.block():
                printer.add_token( self.code )
        else:
            printer.add_token( '@python ' )
            printer.add_token( self.code )

    def __eval__ ( self, context : Context ):
        return PythonNode.execute( context, self.bytecode, self.is_expression )

class PythonContext:
    def __init__ ( self, context : Context ):
        self.context : Context = context
    
    def __getitem__ ( self, key ):
        return self.context.symbols.lookup( key, stop_on = self.context.symbols.prelude, default = KeyError(), raise_default = True )
    
    def __setitem__ ( self, key, value ):
        return self.context.symbols.assign( key, value )
    
    def __contains__ ( self, key ):
        try:
            self[ key ]

            return True
        except KeyError:
            return False

    def get(self, k):
        return self.__getitem__( k )

    def items( self ):
        return iter( () )
    def keys( self ):
        return iter( () )
    def values( self ):
        return iter( () )
    def __len__( self ):
        return 0

    # @overload
    # def get(self, k: _KT, default: Union[_VT_co, _T]) -> Union[_VT_co, _T]: ...
    # def items(self) -> AbstractSet[Tuple[_KT, _VT_co]]: ...
    # def keys(self) -> AbstractSet[_KT]: ...
    # def values(self) -> ValuesView[_VT_co]: ...
    # def __contains__(self, o: object) -> bool: ...


from typing import Any, Callable, List, Optional
from .interval import Interval

ScaleMapper = Callable[[Interval, int, 'Scale'], Any]

class Scale:
    black_keys : 'Scale'
    black_keys_padded : 'Scale'
    white_keys : 'Scale'
    pentatonic_major : 'Scale'

    # major: W W H W W W H
    # minor: W H W W H W W
    @classmethod
    def steps ( cls, steps ):
        notes = [ 0 ]

        acc = 0

        for s in steps:
            acc += s

            notes.append( acc )

        return cls( notes )

    def __init__ ( self, intervals : List[int], mapper : ScaleMapper = None ):
        self.intervals : List[int] = intervals
        self.mapper : Optional[ScaleMapper] = mapper

    def __len__ ( self ):
        return len( self.intervals )

    def __getitem__ ( self, key ) -> Interval:
        return self.interval_at( key )

    def __iter__ ( self ):
        for i in range( len( self ) ):
            yield self[ i ]

    def map ( self, mapper : ScaleMapper ) -> 'Scale':
        return Scale( self.intervals, mapper )

    def interval_at ( self, index : int ) -> Interval:
        l = len( self.intervals )
        
        interval = Interval( octaves = index // l, semitones = self.intervals[ index % l ] )

        if self.mapper is not None:
            return self.mapper( interval, index, self )
        
        return interval

Scale.black_keys = Scale( [ 1, 3, 6, 8, 10 ] )

Scale.black_keys_padded = Scale( [ 1, 3, 3, 6, 8, 10, 10 ] )

Scale.white_keys = Scale( [ 0, 2, 4, 5, 7, 9, 11 ] )

Scale.pentatonic_major = Scale( [ 0, 2, 4, 7, 9 ] )

def inverted ( intervals : List[int], count : int ) -> List[int]:
    if count > 0:
        for i in range( 0, count ):
            intervals[ i ] += 12
    else:
        for i in range( 0, count * -1 ):
            intervals[ i ] -= 12

    return intervals

major_intervals = [ 2, 2, 1, 2, 2, 2, 1 ]

major = Scale.steps( major_intervals )

minor_intervals = [ 2, 1, 2, 2, 1, 2, 2 ]

minor = Scale.steps( minor_intervals )

# Dyads
major_dyad = Scale.steps( [ 4 ] )

minor_dyad = Scale.steps( [ 3 ] )

# Triads
major_triad = Scale.steps( [ 4, 3 ] )

augmented_triad = Scale.steps( [ 4, 4 ] )

minor_triad = Scale.steps( [ 3, 4 ] )

diminished_triad = Scale.steps( [ 3, 3 ] )

# Sevenths
minor_seventh = Scale.steps( [ 3, 4, 3 ] )

major_seventh = Scale.steps( [ 4, 3, 4 ] )

dominant_seventh = Scale.steps( [ 4, 3, 3 ] )

diminished_seventh = Scale.steps( [ 3, 3, 3 ] )

half_diminished_seventh = Scale.steps( [ 3, 3, 4 ] )

minor_major_seventh = Scale.steps( [ 3, 4, 4 ] )

# Perfect Fifth
perfect_fifth = Scale.steps( [ 7 ] )

chords = {
    'm3': minor_dyad,
    'M3': major_dyad,

    'm': minor_triad,
    'M': major_triad,
    'dim': diminished_triad,
    'aug': augmented_triad,
    '+': augmented_triad,

    'm7': minor_seventh,
    'M7': major_seventh,
    'dom7': dominant_seventh,
    '7': dominant_seventh,
    'm7b5': half_diminished_seventh,
    'dim7': diminished_seventh,
    'mM7': minor_major_seventh,

    '5': perfect_fifth
}


from .statement_node import StatementNode
from .statements_list_node import StatementsListNode
from .var_declaration_node import VariableDeclarationStatementNode
from .multi_var_declaration_node import MultiVariableDeclarationStatementNode
from .function_declaration_node import FunctionDeclarationStatementNode
from .for_loop_node import ForLoopStatementNode
from .while_loop_node import WhileLoopStatementNode
from .if_node import IfStatementNode
from .import_node import ImportStatementNode
from .return_node import ReturnStatementNode



from musikla.core import Context, Value, Music
from musikla.core.events import NoteEvent
from musikla.core.theory import Note
from typing import List, Dict, Union, Any

class KeyboardEvent:
    @staticmethod
    def deserialize ( data : str, parameters : Dict[str, Any] ) -> 'KeyboardEvent':
        raise Exception( "Deserialize not implemented" )

    binary : bool = True
    
    def serialize ( self ) -> str:
        raise Exception( "Serialize not implemented" )

    def get_parameters ( self ) -> Dict[str, Any]:
        return {}

class PianoKey(KeyboardEvent):
    @staticmethod
    def deserialize ( data : str, parameters : Dict[str, Any] ) -> 'PianoKey':
        return PianoKey( Note.from_pitch( int( data ) ) )

    def serialize ( self ) -> str:
        return str( int( self.note ) )

    def __init__ ( self, event : Union[Music, NoteEvent, Note, None] ):
        super().__init__()

        if isinstance( event, Music ):
            event = event.first_note( Context() )

        if isinstance( event, NoteEvent ):
            event = event.note

        if event is None:
            raise Exception( "Cannot create PianoKey of None" )

        self.note : Note = event.timeless()
    
    def __eq__ ( self, k ):
        if k is None: return False

        if not isinstance( k, PianoKey ):
            return False

        return self.note == k.note

    def __hash__ ( self ):
        return hash( self.note )

    def __str__ ( self ):
        return str( self.note )

class KeyEvent(KeyboardEvent):
    binary : bool = False

    def __init__ ( self, key = None, value = None, pressed : bool = False ):
        super().__init__()

        self.key = key
        self.value = value
        self.pressed = pressed

    def get_parameters ( self ) -> Dict[str, Any]:
        return { 'key': self.key, 'value': self.value, 'pressed': self.pressed }

    def __hash__ ( self ):
        return hash( '<KeyEvent>' )
    
    def __eq__ ( self, other ):
        if other is None:
            return False

        return isinstance( other, KeyEvent )

class KeyStroke(KeyboardEvent):
    @staticmethod
    def deserialize ( data : str, parameters : Dict[str, Any] ) -> 'KeyStroke':
        return KeyStroke.parse( data )

    def serialize ( self ) -> str:
        return str( self )

    @classmethod
    def parse ( cls, s ):
        parts = s.strip().split( "+" )

        key = parts[ -1 ]

        mods = [ s.strip().lower() for s in parts[ :-1 ] ]

        ctrl = 'ctrl' in mods
        alt = 'alt' in mods
        shift = 'shift' in mods

        return cls( key, ctrl, alt, shift )

    def __init__ ( self, key, ctrl : bool = False, alt : bool = False, shift : bool = False ):
        super().__init__()

        self.ctrl = ctrl
        self.alt = alt
        self.shift = shift
        self.key = key

    def __eq__ ( self, k ):
        if k is None:
            return False

        if not isinstance( k, KeyStroke ):
            return False
        
        return self.ctrl == k.ctrl \
           and self.alt == k.alt \
           and self.shift == k.shift \
           and self.key == k.key

    def __hash__ ( self ):
        return str( self ).__hash__()

    def __str__ ( self ):
        mods = list()

        if self.ctrl: mods.append( 'ctrl' )
        if self.alt: mods.append( 'alt' )
        if self.shift: mods.append( 'shift' )

        mods.append( self.key )

        return '+'.join( [ str(m) for m in mods ] )

class KeyStrokePress( KeyStroke ):
    binary : bool = False

    def __hash__ ( self ):
        return hash( ( "Press ", super().__hash__() ) )

class KeyStrokeRelease( KeyStroke ):
    binary : bool = False
    
    def __hash__ ( self ):
        return hash( ( "Release ", super().__hash__() ) )

class MouseMove( KeyboardEvent ):
    @staticmethod
    def deserialize ( data : str, parameters : Dict[str, Any] ) -> 'MouseMove':
        return MouseMove( parameters[ 'x' ], parameters[ 'y' ] )
    
    def serialize ( self ) -> str:
        return ""

    binary : bool = False

    def __init__ ( self, x : int = 0, y : int = 0 ):
        self.x : int = x
        self.y : int = y
    
    def get_parameters ( self ) -> Dict[str, Any]:
        return { 'x': self.x, 'y': self.y }

    def __hash__ ( self ):
        return hash( '<MouseMove>' )
    
    def __eq__ ( self, other ):
        if other is None:
            return False

        return isinstance( other, MouseMove )

class MouseClick( KeyboardEvent ):
    @staticmethod
    def deserialize ( data : str, parameters : Dict[str, Any] ) -> 'MouseClick':
        return MouseClick(
            parameters[ 'x' ],
            parameters[ 'y' ],
            parameters[ 'button' ],
            parameters[ 'pressed' ]
        )

    def serialize ( self ) -> str:
        return ""

    binary : bool = False

    def __init__ ( self, x : int = 0, y : int = 0, button : int = 0, pressed : bool = False ):
        self.x : int = x
        self.y : int = y
        self.button : int = button
        self.pressed : bool = pressed
    
    def get_parameters ( self ) -> Dict[str, Any]:
        return { 'x': self.x, 'y': self.y, 'button': self.button, 'pressed': self.pressed }

    def __hash__ ( self ):
        return hash( '<MouseClick>' )
    
    def __eq__ ( self, other ):
        if other is None:
            return False

        return isinstance( other, MouseClick )

class MouseScroll( KeyboardEvent ):
    @staticmethod
    def deserialize ( data : str, parameters : Dict[str, Any] ) -> 'MouseScroll':
        return MouseScroll(
            parameters[ 'x' ],
            parameters[ 'y' ],
            parameters[ 'dx' ],
            parameters[ 'dy' ]
        )
    
    def serialize ( self ) -> str:
        return ""

    binary : bool = False

    def __init__ ( self, x : int = 0, y : int = 0, dx : int = 0, dy : int = 0 ):
        self.x : int = x
        self.y : int = y
        self.dx : int = dx
        self.dy : int = dy
    
    def get_parameters ( self ) -> Dict[str, Any]:
        return { 'x': self.x, 'y': self.y, 'dx': self.dx, 'dy': self.dy }

    def __hash__ ( self ):
        return hash( '<MouseScroll>' )
    
    def __eq__ ( self, other ):
        if other is None:
            return False

        return isinstance( other, MouseScroll )

class EventSource():
    def listen ( self ):
        pass

    def close ( self ):
        pass


from .context_modifier_node import ContextModifierNode
from .length_modifier_node import LengthModifierNode
from .octave_modifier_node import OctaveModifierNode
from .signature_modifier_node import SignatureModifierNode
from .tempo_modifier_node import TempoModifierNode
from .velocity_modifier_node import VelocityModifierNode
from .instrument_modifier_node import InstrumentModifierNode

from .block_context_modifier_node import BlockContextModifierNode
from .voice_block_modifier import VoiceBlockModifier



from typing import Tuple, Union
from .instrument import Instrument, GeneralMidi
from fractions import Fraction
from copy import copy

class Voice:
    unknown : 'Voice' = None

    def __init__ ( self, 
                   name : str = 'default', 
                   instrument : Instrument = None,
                   time_signature : Tuple[int, int] = (4, 4),
                   velocity : int = 127,
                   octave : int = 4,
                   value : float = 1,
                   tempo : int = 60
                 ):
        self.id : int = id( self )
        self.name : str = name
        self.instrument : Instrument = instrument or Instrument( 'Acoustic Grand Piano', GeneralMidi.AcousticGrandPiano )
        self.time_signature : Tuple[int, int] = time_signature
        self.velocity : int = velocity
        self.octave : int = octave
        self.value : float = value
        self.tempo : int = tempo

    def clone ( self,
                name : str = None, 
                instrument : Instrument = None,
                time_signature : Tuple[int, int] = None,
                velocity : int = None,
                octave : int = None,
                value : float = None,
                tempo : int = None ):
        return Voice(
            name = self.name if name is None else name,
            instrument = self.instrument if instrument is None else instrument,
            time_signature = self.time_signature if time_signature is None else time_signature,
            velocity = self.velocity if velocity is None else velocity,
            octave = self.octave if octave is None else octave,
            value = self.value if value is None else value,
            tempo = self.tempo if tempo is None else tempo
        )

    def revoice ( self, event ):
        if event.voice == self:
            return event

        event = event.clone()

        if self.octave != event.voice.octave:
            event.octave = event.octave - event.voice.octave + self.octave

        if self.velocity != event.voice.velocity:
            event.velocity = self.velocity

        event.voice = self

        return event

    def get_beats_per_bar ( self ) -> int:
        return self.time_signature[ 0 ]

    def get_beat_value_absolute ( self ) -> Fraction:
        ( u, l ) = self.time_signature

        if u >= 6 and u % 3 == 0:
            return Fraction( 3, 2 * l )
        else:
            return Fraction( 1, l )

    def get_bar_value_absolute ( self ) -> Fraction:
        """Returns how long a measure should be, in note fractions"""
        return self.get_beat_value_absolute() * self.get_beats_per_bar()

    def get_bar_duration_absolute ( self ) -> int:
        """Returns how long a measure should be, in milliseconds"""
        return self.get_duration_absolute( self.get_bar_value_absolute() )

    def get_value ( self, value : float = None ) -> float:
        if value == None:
            return self.value
        else:
            return self.value * value

    def get_relative_value ( self, value : Union[float, Fraction] ) -> Fraction:
        if not isinstance( value, Fraction ):
            value = Fraction( value )

        return value / self.value

    def get_duration_ratio ( self ) -> float:
        ( u, l ) = self.time_signature

        if u >= 6 and u % 3 == 0:
            return 1.5 / l
        else:
            return 1 / l

    def get_duration ( self, value : float = None ) -> int:
        """Transform a not value into the real world milliseconds it takes, according to the voice's tempo and time signature"""
        return self.get_duration_absolute( self.get_value( value ) )

    def get_beat_duration ( self ) -> int:
        return int( 60 / self.tempo )

    def get_duration_absolute ( self, value : float = None ) -> int:
        beat_duration = 60 / self.tempo

        whole_note_duration = beat_duration * 1000.0 / self.get_duration_ratio()

        return int( whole_note_duration * value )

    def from_duration ( self, milliseconds : int, max_denominator : int = 32 ) -> Fraction:
        """Transform a duration in milliseconds to an approximated note value relative to the tempo and time signature and default not length"""
        return self.get_relative_value( self.from_duration_absolute( milliseconds, max_denominator ) )
        
    def from_duration_absolute ( self, milliseconds : int, max_denominator : int = 32 ) -> Fraction:
        """Transform a duration in milliseconds to an approximated note value relative to the tempo and time signature"""
        beat_duration = 60 / self.tempo

        whole_note_duration = beat_duration * 1000.0 / self.get_duration_ratio()
        
        return Fraction( milliseconds / whole_note_duration ).limit_denominator( max_denominator )

    def __eq__ ( self, other ):
        if not isinstance( other, Voice ): return False

        return  self.name == other.name \
            and self.instrument == other.instrument \
            and self.time_signature == other.time_signature \
            and self.velocity == other.velocity \
            and self.octave == other.octave \
            and self.value == other.value \
            and self.tempo == other.tempo

Voice.unknown = Voice( "(unknown)", None )



from typing import Optional
from .event import VoiceEvent
from ..voice import Voice

BAR_STANDARD = 0
BAR_DOUBLE = 1
BAR_END = 2
BAR_BEGIN_REPEAT = 3
BAR_END_REPEAT = 4
BAR_BEGIN_END_REPEAT = 5

class BarNotationEvent ( VoiceEvent ):
    def __init__ ( self, timestamp : int, voice : Voice, staff : Optional[int] = None, kind : int = BAR_STANDARD ):
        super().__init__( timestamp, voice, staff )

        self.kind : int = kind
            
    def __str__ ( self ) -> str:
        return f"[newbar]"

class StaffNotationEvent ( VoiceEvent ):
    def __init__ ( self, timestamp : int, voice : Voice, staff : Optional[int] = None, kind : int = BAR_STANDARD ):
        super().__init__( timestamp, voice, staff )

        self.kind : int = kind

    def __str__ ( self ) -> str:
        return f"[newstaff]"




from .value import Value, CallableValue
from .context import Context
from .symbols_scope import Ref
from musikla.parser.abstract_syntax_tree.node import Node
from musikla.parser.abstract_syntax_tree.expressions.string_literal_node import StringLiteralNode
from musikla.parser.abstract_syntax_tree.expressions.variable_expression_node import VariableExpressionNode

from typing import Union, Optional, _GenericAlias, Dict
from inspect import signature, Signature, Parameter
from typeguard import check_type

def is_type_of ( typehint, value ) -> bool:
    if value == typehint:
        return True
    elif isinstance( typehint, _GenericAlias ):
        origin = typehint.__origin__

        if origin == Union or origin == Optional:
            return any( is_type_of( var, value ) for var in typehint.__args__ )
        else:
            return False
    else:
        return False

class CallablePythonValue(CallableValue):
    @staticmethod
    def call ( fn, context, args = [], kargs = {} ):
        if isinstance( fn, CallableValue ):
            return fn( context, args, kargs )
        else:
            return CallablePythonValue( fn )( context, args, kargs )

    @staticmethod
    def raw_call ( callable, context ):
        def closure ( *args, **kargs ):
            nonlocal context

            literal_args = [ StringLiteralNode( value ) for value in args ]
            literal_kargs = dict( ( key, StringLiteralNode( value ) ) for key, value in kargs.items() )

            return callable( context, literal_args, literal_kargs )

        return closure

    def __init__ ( self, value ):
        try:
            self.signature : Signature = signature( value )
        except ValueError:
            self.signature : Signature = None

        self.callable = value

        super().__init__( self.wrapper )

    def raw ( self ):
        return self.callable

    def eval_argument ( self, context : Context, parameter : Parameter, node : Node, arg_name : str ):
        if is_type_of( parameter.annotation, Node ):
            return node

        if is_type_of( parameter.annotation, Ref ):
            check_type( arg_name, node, VariableExpressionNode )

            pointer = context.symbols.pointer( node.name )

            return Ref( pointer )

        value = Value.eval( context.fork(), node )

        if parameter.annotation != Parameter.empty:
            check_type( arg_name, value, parameter.annotation )
        
        if isinstance( value, CallableValue ):
            return CallablePythonValue.raw_call( value, context )

        return value

    def eval_var_keyword ( self, context : Context, parameter : Parameter, nodes : Dict[ str, Node ] ):
        return dict( ( key, self.eval_argument( context, parameter, value, key ) ) for key, value in nodes.items() )

    def wrapper ( self, context, *args, **kargs ):
        args_values = list()

        kargs_values = dict()

        if self.signature != None:
            pass_context : bool = False

            i = 0

            for arg_name in self.signature.parameters:
                parameter = self.signature.parameters[ arg_name ]

                if i == 0 and not pass_context and is_type_of( parameter.annotation, Context ):
                    pass_context = True

                    args_values.append( context )

                    continue

                # Let's treat this as positional arguments
                if parameter.kind == Parameter.VAR_POSITIONAL:
                    while i < len( args ):
                        args_values.append( self.eval_argument( context, parameter, args[ i ], arg_name ) )

                        i = i + 1
                elif i < len( args ):
                    args_values.append( self.eval_argument( context, parameter, args[ i ], arg_name ) )

                    i = i + 1
                # Treat this as keywork arguments~
                elif parameter.kind == Parameter.VAR_KEYWORD:
                    kargs_values.update( self.eval_var_keyword( context, parameter, kargs ) )
                else:
                    if arg_name in kargs:
                        kargs_values[ arg_name ] = self.eval_argument( context, parameter, kargs[ arg_name ], arg_name )
        else:
            args_values = [ Value.eval( context.fork(), node ) for node in args ]

            kargs_values = dict( [ ( key, Value.eval( context.fork(), node ) ) for key, node in kargs.items() ] )

        return self.callable( *args_values, **kargs_values )



from typing import Tuple
from ..statements import StatementNode, VariableDeclarationStatementNode
from ..expressions import FunctionExpressionNode, StringLiteralNode, VariableExpressionNode
from .macro import MacroNode
from musikla.core import Instrument

class VoiceDeclarationMacroNode( MacroNode ):
    def __init__ ( self, name, modifiers = None, parent = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.name = name

        kargs = {}

        if modifiers != None:
            kargs[ 'modifiers' ] = modifiers
        
        if parent != None:
            kargs[ 'inherit' ] = parent
 
        self.virtual_node = VariableDeclarationStatementNode(
            self.name,
            FunctionExpressionNode( VariableExpressionNode( "voices\\create" ), [ StringLiteralNode( name ) ], kargs )
        )



from musikla.parser.printer import CodePrinter
from musikla.core.context import StackFrame
from ..node import Node
from .statement_node import StatementNode
from typing import Tuple, Optional
from musikla.core import Value, Context

class ReturnStatementNode( StatementNode ):
    def __init__ ( self, expression : Node = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.expression : Optional[Node] = expression

    def __eval__ ( self, context : Context ):
        stack_frame : Optional[StackFrame] = context.symbols.lookup( 'stack_frame', container = 'stack' )

        if stack_frame is None:
            raise Exception( "Cannot return here, no stack frame found." )

        if self.expression is None:
            stack_frame.ret()
        else:
            val = Value.eval( context, self.expression )

            stack_frame.ret( Value.assignment( val ) )

            return val

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( "return " )

        if self.expression is not None:
            self.expression.to_source( printer )



from .transformer import Transformer
from ..event import MusicEvent
from ..note import NoteEvent, NoteOnEvent, NoteOffEvent
from ..chord import ChordEvent
from ...music import MusicBuffer
from collections import defaultdict
from typing import Callable, Dict, List, Optional, Tuple

class ComposeChordsTransformer( Transformer ):
    def __init__ ( self ):
        super().__init__()
        
        self.buffer : List[NoteEvent] = []
        self.last_timestamp : int = 0

    def flush_buffer ( self ):
        if not self.buffer:
            return
        
        if len( self.buffer ) == 1:
            self.add_output( self.buffer[ 0 ] )
        else:
            # We know that the first event in the buffer must be a note event
            # because we only append to the buffer if the event is a note event
            # or if the buffer already has any events in it
            first_name = self.buffer[ 0 ].voice.name
            
            # A variable that is true if all the note events in the buffer belong
            # to the same voice
            single_voice = all( ( ev.voice.name == first_name for ev in self.buffer if isinstance( ev, NoteEvent ) ) )

            per_voice : Dict[Tuple[str, Optional[int]], List[NoteEvent]] = {}

            voiceless : List[MusicEvent] = []

            for ev in self.buffer:
                if isinstance( ev, NoteEvent ) and ev.voice is not None:
                    voice_name = ev.voice.name

                    if voice_name not in per_voice:
                        per_voice[ ( voice_name, ev.staff ) ] = [ ev ]
                    else:
                        per_voice[ ( voice_name, ev.staff ) ].append( ev )
                else:
                    voiceless.append( ev )

            # TODO Handle notes played at the same time for the same voice
            # but with different durations
            for notes in per_voice.values():
                if len( notes ) <= 2:
                    for note in notes:
                        self.add_output( note )
                else:
                    self.add_output( ChordEvent(
                        timestamp = notes[ 0 ].timestamp,
                        pitches = [ int( n ) for n in notes ],
                        duration = max( ( n.duration for n in notes ) ),
                        value = max( ( n.value for n in notes ) ),
                        voice = notes[ 0 ].voice,
                        tied = notes[ 0 ].tied,
                        velocity = notes[ 0 ].velocity
                    ) )

            for ev in voiceless:
                self.add_output( ev )

        self.buffer.clear()

    def transform ( self ):
        while True:
            done, event = yield

            if done: break

            if event.timestamp != self.last_timestamp:
                self.flush_buffer()
            
            if isinstance( event, NoteEvent ) or self.buffer:
                self.buffer.append( event )
            else:
                self.add_output( event ) 
            
            self.last_timestamp = event.timestamp

        self.flush_buffer()


from musikla.core.events.event import VoiceEvent
from musikla.core.events.transformers import ComposeNotesTransformer, ComposeChordsTransformer, VoiceIdentifierTransformer, AnnotateTransformer, EnsureOrderTransformer
from musikla.core.events import MusicEvent
from .sequencer import Sequencer, SequencerFactory
from typing import List, Optional
from pathlib import Path
import json

class DebugSequencer ( Sequencer ):
    def __init__ ( self, filename : str ):
        super().__init__()

        self.filename : str = filename

        self.events : List[MusicEvent] = []
        
        self.set_transformers(
            ComposeNotesTransformer(),
            ComposeChordsTransformer(),
            VoiceIdentifierTransformer(),
            AnnotateTransformer()
        )
    
    @property
    def playing ( self ):
        return True

    def build ( self ) -> str:
        items = list()
        groups = list()
        # groups.append( {
        #     'id': 'default',
        #     'content': 'Default'
        # } )
        # groups.append( {
        #     'id': '<voiceless>',
        #     'content': '<Voiceless>'
        # } )

        for i, event in enumerate( self.events ):
            voice_name = "<voiceless>"

            if isinstance( event, VoiceEvent ) and event.voice is not None:
                voice_name = event.voice.name

            group_name = f"{voice_name}: { event.staff }"

            if not any( g[ 'id' ] == group_name for g in groups ):
                groups.append( { 'id': group_name, 'content': group_name } )

            obj = {
                'id': i,
                'content': str( event ),
                'title': str( event ),
                'start': event.timestamp,
                'group': group_name,
                'type': 'point'
            }

            if event.end_timestamp > event.timestamp:
                obj[ 'type'] = 'range'
                obj[ 'end' ] = event.end_timestamp

            items.append( obj )

        template = f"""<html>
    <head>
        <script type="text/javascript" src="http://unpkg.com/vis-timeline@latest/standalone/umd/vis-timeline-graph2d.min.js"></script>
        <link href="http://unpkg.com/vis-timeline@latest/styles/vis-timeline-graph2d.min.css" rel="stylesheet" type="text/css" />
        <style type="text/css">
            #visualization {{
                width: 100%;
                height: 800px;
                border: 1px solid lightgray;
            }}
        </style>
    </head>
    <body>
        <div id="visualization"></div>
        <div id="details"></div>
        <script type="text/javascript">
        // DOM element where the Timeline will be attached
        var container = document.getElementById('visualization');

        var detailsContainer = document.getElementById('details');

        // Create a DataSet (allows two way data-binding)
        var groups = { json.dumps( groups ) }

        var items = new vis.DataSet( { json.dumps( items ) } );

        // Configuration for the Timeline
        var options = {{
            multiselect: true,
            groupHeightMode: 'fixed'
        }};

        // Create a Timeline
        var timeline = new vis.Timeline(container, items, groups, options);

        timeline.on( 'select', e => {{
            if ( e.items.length == 0 ) {{
                detailsContainer.innerHTML = '';
            }} else {{
                detailsContainer.innerHTML = e.items
                    .map( id => items.get( id ) )
                    .map( ev => "<p>" + ev.content + " [" + ev.start + " - " + ev.end + " = " + ( ev.end - ev.start ) + "]" + "</p>" )
                    .join( '' );
            }}
        }} );
        </script>
    </body>
</html>
"""

        return template

    def get_time ( self ):
        return 0

    def on_event ( self, event : MusicEvent ):
        self.events.append( event )
    
    def on_close ( self ):
        with open( self.filename, 'w' ) as f:    
            f.write( self.build() )
                
            f.flush()

    def join ( self ):
        pass

    def start ( self ):
        pass

class DebugSequencerFactory( SequencerFactory ):
    def init ( self ):
        self.name = 'debug'
    
    def from_str ( self, uri : str, args ) -> Optional[DebugSequencer]:
        suffix = ( Path( uri ).suffix or '' ).lower()

        if suffix == '.html' or suffix == '.htm':
            return DebugSequencer( uri )
        



from musikla.parser.printer import CodePrinter
from .music_node import MusicNode
from .music_parallel_node import MusicParallelNode
from musikla.core.theory import NoteAccidental, NotePitchClasses, NotePitchClassesInv, Note
from musikla.core.events import NoteEvent
from typing import List, Tuple

class NoteNode( MusicNode ):
    def __init__ ( self, note : Note, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.note : Note = note
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( self.note.to_string( base_octave = 0 ) )

    def get_events ( self, context ):
        note = NoteEvent(
            timestamp = context.cursor,
            pitch_class = self.note.pitch_class,
            duration = context.get_duration( self.note.value ),
            value = context.get_value( self.note.value ),
            octave = context.voice.octave + ( self.note.octave or 0 ),
            voice = context.voice,
            velocity = context.voice.velocity,
            accidental = self.note.accidental,
            staff = 0
        )

        context.cursor += note.duration

        yield note

    def as_chord ( self, intervals : List[int] ) -> MusicParallelNode:
        notes = self.note.as_chord( intervals ).to_notes()

        nodes = [ NoteNode( note, self.position ) for note in notes ]

        return MusicParallelNode( nodes, position = self.position )



from musikla.parser.printer import CodePrinter
from typing import Tuple
from musikla.core import Context
from ..node import Node

class PropertyAccessorNode( Node ):
    def __init__ ( self, expression : Node, name : Node, as_attr : bool, position : Tuple[int, int, int] = None ):
        super().__init__( position )
        
        self.expression : Node = expression
        self.name : Node = name
        self.as_attr : bool = as_attr

    def assign ( self, context, value, local : bool = False ):
        expr = self.expression.eval( context )
        name = self.name.eval( context )

        if self.as_attr:
            setattr( expr, name, value )
        else:
            expr[ name ] = value
        
    def lookup_assign ( self, context, value_fn, local : bool = False ):
        expr = self.expression.eval( context )
        name = self.name.eval( context )

        if self.as_attr:
            value = getattr( expr, name, None )
        else:
            value = expr[ name ]

        value = value_fn( value )

        if self.as_attr:
            setattr( expr, name, value )
        else:
            expr[ name ] = value

    def __eval__ ( self, context : Context ):
        expr = self.expression.eval( context )
        name = self.name.eval( context )

        if self.as_attr:
            return getattr( expr, name, None )
        else:
            return expr[ name ]

    def to_source ( self, printer : CodePrinter ):
        from .string_literal_node import StringLiteralNode

        self.expression.to_source( printer )

        if self.as_attr and isinstance( self.name, StringLiteralNode ):
            printer.add_token( '::' + self.name.value )
        else:
            with printer.block( '::[', ']' ):
                self.name.to_source( printer )


from typing import Optional, Union

class Instrument():
    @staticmethod
    def from_program ( program : int, bank : int = None, soundfont : Union[int, str] = None ) -> 'Instrument':
        return Instrument( GeneralMidi.get_name( program ), program, bank, soundfont )

    def __init__ ( self, name : str, program : int, bank : int = None, soundfont : Union[int, str] = None ):
        self.name : str = name
        self.program : int = program
        self.bank : Optional[int] = bank
        self.soundfont : Optional[Union[int, str]] = soundfont

class GeneralMidi():
    @staticmethod
    def get_name ( program : int ) -> str:
        for attr in dir( program ):
            if attr.startswith( "_" ): continue

            if hasattr( GeneralMidi, attr ) and getattr( GeneralMidi, attr ) == program:
                return attr
        
        return "(Unknown)"

    # Piano
    AcousticGrandPiano = 1
    BrightAcousticPiano = 2
    ElectricGrandPiano = 3
    HonkyTonkPiano = 4
    ElectricPiano1 = 5
    ElectricPiano2 = 6
    Harpsichord = 7
    Clavinet = 8

    # Chromatic Percussion
    Celesta = 9
    Glockenspiel = 10
    MusicBox = 11
    Vibraphone = 12
    Marimba = 13
    Xylophone = 14
    TubularBells = 15
    Dulcimer = 16

    # Organ
    DrawbarOrgan = 17
    PercussiveOrgan = 18
    RockOrgan = 19
    ChurchOrgan = 20
    ReedOrgan = 21
    Accordion = 22
    Harmonica = 23
    TangoAccordion = 24

    # Guitar
    AcousticGuitarNylon = 25
    AcousticGuitarSteel = 26
    ElectricGuitarJazz = 27
    ElectricGuitarClean = 28
    ElectricGuitarMuted = 29
    OverdrivenGuitar = 30
    DistortionGuitar = 31
    GuitarHarmonics = 32

    # Bass
    AcousticBass = 33
    ElectricBassFinger = 34
    ElectricBassPick = 35
    FretlessBass = 36
    SlapBass1 = 37
    SlapBass2 = 38
    SynthBass1 = 39
    SynthBass2 = 40

    # Strings
    Violin = 41
    Viola = 42
    Cello = 43
    Contrabass = 44
    TremoloStrings = 45
    PizzicatoStrings = 46
    OrchestralHarp = 47
    Timpani = 48

    # Ensemble
    StringEnsemble1 = 49
    StringEnsemble2 = 50
    SynthStrings1 = 51
    SynthStrings2 = 52
    ChoirAahs = 53
    VoiceOohs = 54
    SynthChoir = 55
    OrchestraHit = 56

    # Brass
    Trumpet = 57
    Trombone = 58
    Tuba = 59
    MutedTrumpet = 60
    FrenchHorn = 61
    BrassSection = 62
    SynthBrass1 = 63
    SynthBrass2 = 64

    # Reed
    SopranoSax = 65
    AltoSax = 66
    TenorSax = 67
    BaritoneSax = 68
    Oboe = 69
    EnglishHorn = 70
    Bassoon = 71
    Clarinet = 72

    # Pipe
    Piccolo = 73
    Flute = 74
    Recorder = 75
    PanFlute = 76
    BlownBottle = 77
    Shakuhachi = 78
    Whistle = 79
    Ocarina = 80

    # Synth Lead
    Lead1Square = 81
    Lead2Sawtooth = 82
    Lead3Calliope = 83
    Lead4Chiff = 84
    Lead5Charang = 85
    Lead6Vvoice = 86
    Lead7Fifths = 87
    Lead8BassLead = 88

    # Synth Pad
    Pad1NewAge = 89
    Pad2Warm = 90
    Pad3Polysynth = 91
    Pad4Choir = 92
    Pad5Bowed = 93
    Pad6Metallic = 94
    Pad7Halo = 95
    Pad8Sweep = 96

    # Synth Effects
    FX1Rain = 97
    FX2Soundtrack = 98
    FX3Crystal = 99
    FX4Atmosphere = 100
    FX5Brightness = 101
    FX6Goblins = 102
    FX7Echoes = 103
    FX8SciFi = 104

    # Ethnic
    Sitar = 105
    Banjo = 106
    Shamisen = 107
    Koto = 108
    Kalimba = 109
    Bagpipe = 110
    Fiddle = 111
    Shanai = 112

    # Percussive
    TinkleBell = 113
    Agogo = 114
    SteelDrums = 115
    Woodblock = 116
    TaikoDrum = 117
    MelodicTom = 118
    SynthDrum = 119
    ReverseCymbal = 120

    # Sound effects
    GuitarFretNoise = 121
    BreathNoise = 122
    Seashore = 123
    BirdTweet = 124
    TelephoneRing = 125
    Helicopter = 126
    Applause = 127
    Gunshot = 128



from musikla.parser.printer import CodePrinter
from typing import Tuple
from .music_node import MusicSequenceBase

class MusicSequenceNode( MusicSequenceBase ):
    def __init__ ( self, nodes, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.expressions = nodes

    def to_source ( self, printer : CodePrinter ):
        for i in range( len( self.expressions ) ):
            if i > 0: printer.add_token( ' ' )

            self.expressions[ i ].to_source( printer )

    def values ( self, context ):
        for node in self.expressions:
            yield node.eval( context )

    def check_returns(self):
        return False

    def __iter__ ( self ):
        return iter( self.expressions )



#!/usr/bin/python3
#!/usr/bin/python3

from sys import argv
import asyncio

def main():
    try:
        from musikla.cli_application import CliApplication

        asyncio.run( CliApplication( argv[ 1: ] ).run() )
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()


from collections import defaultdict
from musikla.parser.printer import CodePrinter
from typing import Dict, Generic, List, Tuple, TypeVar
from musikla.core import Music
from .music_node import MusicNode
from musikla.core import merge_sorted

T = TypeVar('T')

class Box( Generic[T] ):
    def __init__ ( self, value : T ):
        self.value : T = value
    
    def gset ( self, new_value : T ) -> T:
        old_value : T = self.value

        self.value = new_value

        return old_value
    
    def set ( self, value : T ):
        self.value = value

class MusicParallelNode( MusicNode ):
    def __init__ ( self, nodes, position : Tuple[int, int, int] = None ):
        super().__init__( position )
        
        self.expressions = nodes
    
    def to_source ( self, printer : CodePrinter ):
        for i in range( len( self.expressions ) ):
            if i > 0: printer.add_token( ' | ' )

            self.expressions[ i ].to_source( printer )

    def fork_and_get_events ( self, node, forks, context, staff_count : Box[int] ):
        forked = context.fork()

        forks.append( forked )

        value = node.eval( forked )

        def _update_staff ( events ):
            nonlocal staff_count
            
            staff_map : Dict[int, int] = defaultdict( lambda: staff_count.gset( staff_count.value + 1 ) )

            for event in events:
                if event.staff is not None:
                    event.staff = staff_map[ event.staff ]

                yield event

        if isinstance( value, Music ):
            return _update_staff( value.expand( forked ) )

        return iter(())

    def get_events ( self, context ):
        forks = []

        staff_count : Box[int] = Box( 0 )

        notes = map( lambda node: self.fork_and_get_events( node, forks, context, staff_count ), self.expressions )
        notes = merge_sorted( notes, lambda note: note.timestamp )

        try:
            for note in notes:
                yield note
        finally:
            context.join( *forks )
        


from musikla.core.events import MusicEvent, VoiceEvent
from .transformer import Transformer
from typing import List, Dict, Union

class EnsureOrderTransformer(Transformer):
    def __init__ ( self, label : str = None, total_order : bool = True ):
        super().__init__()
        
        self.label : str = label
        self.total_order : bool = total_order

        self.last_timestamp : int = 0
        self.voices : Dict[str, int] = {}

    def print_warning ( self, event : MusicEvent ):
        print( f"{self.label}: { event.timestamp } < { self.last_timestamp } { event.__dict__ }" )

    def print_warning_voice ( self, event : MusicEvent ):
        print( f"{self.label} [{ event.voice.name }]: { event.timestamp } < { self.last_timestamp } { event.__dict__ }" )

    def transform ( self ):
        while True:
            done, event = yield

            if done: break

            if self.total_order or not isinstance( event, VoiceEvent ):
                if event.timestamp < self.last_timestamp:
                    self.print_warning( event )
                else:
                    self.last_timestamp = event.timestamp
            else:
                if event.voice.name not in self.voices:
                    self.voices[ event.voice.name ] = 0
                
                if event.timestamp < self.voices[ event.voice.name ]:
                    self.print_warning_voice( event )
                else:
                    self.voices[ event.voice.name ] = event.timestamp
            
            self.add_output( event )
        
        self.end_output()





from musikla.libraries.keyboard.buffer import KeyboardBuffer
from musikla.libraries.keyboard.virtual_player import VirtualPlayer
from musikla.core import Context, Library, Music, Value
from musikla.audio import Player
from musikla.core.callable_python_value import CallablePythonValue
from typing import Dict, List, Set, Type, Union, Optional, Iterator, Tuple, Any, cast
from musikla.parser.abstract_syntax_tree import Node
from asyncio import Future, sleep, create_task
from io import FileIO
from pathlib import Path
from .grid import Grid
from .keyboard import Keyboard
from .event import EventSource, KeyEvent, KeyStroke, KeyStrokePress, KeyStrokeRelease, KeyboardEvent, MouseClick, MouseMove, MouseScroll, PianoKey
from .action import KeyAction

def register_key ( context : Context, keyboard : Keyboard, key : Node, expression : Node, args : List[str] = [], toggle : Node = None, hold : Node = None, repeat : Node = None, extend : Node = None, release : Node = None ):
    return keyboard.register_key( context, key, expression, args, toggle, hold, repeat, extend, release )

def push_flags ( context : Context, keyboard : Keyboard, *flags : Node ):
    return keyboard.push_flags( context, *flags )

def pop_flags ( context : Context, keyboard : Keyboard, *flags : Node ):
    return keyboard.pop_flags( context, *flags )

def push_prefix ( context : Context, keyboard : Keyboard, expression : Node ):
    return keyboard.push_prefix( context, expression )

def pop_prefix ( context : Context, keyboard : Keyboard ):
    return keyboard.pop_prefix( context )

def start ( context : Context, keyboard : Keyboard, key : Union[str, KeyboardEvent] ):
    return keyboard.start( key )

def stop ( context : Context, keyboard : Keyboard, key : Union[str, KeyboardEvent] ):
    return keyboard.stop( key )

def start_all ( context : Context, keyboard : Keyboard ):
    return keyboard.start_all()

def stop_all ( context : Context, keyboard : Keyboard ):
    return keyboard.stop_all()

def on_press ( context : Context, key : Union[str, KeyboardEvent] ):
    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    return lib.on_press( key )

def on_release ( context : Context, key : Union[str, KeyboardEvent] ):
    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    return lib.on_release( key )

def keyboard_create ( context : Context ) -> Keyboard:
    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    return lib.create()

def keyboard_close ( context : Context, keyboard : Keyboard ):
    if keyboard != None:
        keyboard.close()
    else:
        lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

        lib.close()

def keyboard_record ( context : Context, file : str ):
    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    lib.record( file )

def keyboard_replay ( context : Context, file : str ):
    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    lib.replay( file )

def keyboard_open_repl ( context : Context ):
    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    lib.eval( context )

    return None

def keyboard_readperf ( context : Context, file : str, keyboards : List[Keyboard] = None ) -> Music:
    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    return lib.readperf( context, file, keyboards )

def keyboard_filedialog ( context : Context, title : str = "File", text : str = "File Path:", default_value : str = None, cb = None ):
    from prompt_toolkit.patch_stdout import patch_stdout
    from prompt_toolkit.shortcuts import input_dialog
    from prompt_toolkit.completion import PathCompleter

    lib : KeyboardLibrary = cast( KeyboardLibrary, context.library( KeyboardLibrary ) )

    def _run ():
        with patch_stdout():
            application = input_dialog( title = title, text = text, completer = PathCompleter() )

            with application.input.raw_mode():
                application.input.read_keys()

            application.layout.current_control.buffer.insert_text(default_value or "")
            
            return application.run_async()

    lib.prompt( _run, cb )

class KeyboardLibrary(Library):
    def __init__ ( self, player : Player ):
        super().__init__( "keyboard" )

        self.player : Player = player
        self.paused : bool = False
    
    def on_link ( self, script ):
        self.assign_internal( "keyboards", list() )
        self.assign_internal( "event_sources", list() )
        self.assign_internal( "event_types", dict() )
        
        self.assign( "create", CallablePythonValue( keyboard_create ) )
        self.assign( "push_flags", CallablePythonValue( push_flags ) )
        self.assign( "pop_flags", CallablePythonValue( pop_flags ) )
        self.assign( "push_prefix", CallablePythonValue( push_prefix ) )
        self.assign( "pop_prefix", CallablePythonValue( pop_prefix ) )
        self.assign( "register", CallablePythonValue( register_key ) )
        self.assign( "on_press", CallablePythonValue( on_press ) )
        self.assign( "on_release", CallablePythonValue( on_release ) )
        self.assign( "start", CallablePythonValue( start ) )
        self.assign( "stop", CallablePythonValue( stop ) )
        self.assign( "start_all", CallablePythonValue( start_all ) )
        self.assign( "stop_all", CallablePythonValue( stop_all ) )
        self.assign( "close", CallablePythonValue( keyboard_close ) )

        self.assign( "record", CallablePythonValue( keyboard_record ) )
        self.assign( "readperf", CallablePythonValue( keyboard_readperf ) )
        self.assign( "replay", CallablePythonValue( keyboard_replay ) )
        self.assign( "open_repl", CallablePythonValue( keyboard_open_repl ) )
        self.assign( "filedialog", CallablePythonValue( keyboard_filedialog ) )

        self.assign( "Grid", Grid )
        self.assign( "Buffer", KeyboardBuffer )

        self.assign( "KeyEvent", KeyEvent )
        self.assign( "KeyStroke", KeyStroke )
        self.assign( "KeyStrokePress", KeyStrokePress )
        self.assign( "KeyStrokeRelease", KeyStrokeRelease )
        self.assign( "PianoKey", PianoKey )
        self.assign( "MouseClick", MouseClick )
        self.assign( "MouseMove", MouseMove )
        self.assign( "MouseScroll", MouseScroll )

        self.register_event_type(
            PianoKey, KeyStroke, MouseClick, MouseMove, MouseScroll
        )

        self.context.script.import_module( self.context, Path( __file__ ).parent / "library.mkl", save_cache = False )
        
    def register_event_type ( self, *events : Type[KeyboardEvent] ):
        for event in events:
            self.event_types[ event.__name__ ] = event

    @property
    def keyboards ( self ) -> List[Keyboard]:
        return self.lookup_internal( "keyboards" )

    @property
    def record_file ( self ) -> str:
        return self.lookup_internal( 'record_file' )
    
    @property
    def record_fd ( self ) -> FileIO:
        return self.lookup_internal( 'record_fd' )
        
    @property
    def record_start ( self ) -> int:
        return self.lookup_internal( 'record_start' )

    @property
    def event_sources ( self ) -> List[EventSource]:
        return self.lookup_internal( 'event_sources' )

    @property
    def event_types ( self ) -> Dict[str, Type[KeyboardEvent]]:
        return self.lookup_internal( 'event_types' )

    @property
    def keys ( self ) -> Iterator[Tuple[KeyboardEvent, KeyAction]]:
        for kb in self.keyboards:
            for key, action in kb.keys.items():
                yield key, action

    @property
    def actions ( self ) -> Iterator[KeyAction]:
        return ( action for key, action in self.keys )

    @property
    def pressed ( self ) -> Iterator[KeyAction]:
        return ( action for action in self.actions if action.is_playing )

    @property
    def pressed_keys ( self ) -> Iterator[KeyboardEvent]:
        return ( action.key for action in self.pressed )
    
    @property
    def has_keys ( self ) -> bool:
        iter = self.keys

        has_keys = next( self.keys, None ) != None

        if hasattr( iter, 'close' ):
            cast( Any, iter ).close()

        return has_keys

    def create ( self ) -> Keyboard:
        keyboard = Keyboard( self.context )

        self.keyboards.append( keyboard )

        return keyboard
    
    def close ( self, keyboard : Optional[Keyboard] = None ):
        self.lookup_internal( "keyboards" )

        if keyboard != None:
            if keyboard in self.keyboards:
                self.keyboards.remove( keyboard )
        else:
            for kb in self.keyboards: 
                kb.close( closing = True )
        
            self.keyboards.clear()

            fd = self.record_fd

            if fd is not None:
                fd.close()

        if not self.has_keys:
            close_future : Future = self.lookup_internal( "close_future" )
            
            if close_future != None:
                close_future.set_result( None )

    def start ( self, key : KeyboardEvent ):
        for kb in self.keyboards:
            kb.start( key )
        
    def stop ( self, key : KeyboardEvent ):
        for kb in self.keyboards:
            kb.stop( key )

    def start_all ( self ):
        for kb in self.keyboards:
            kb.start_all()
        
    def stop_all ( self ):
        for kb in self.keyboards:
            kb.stop_all()

    def on_press ( self, key : Union[str, KeyboardEvent] ):
        if self.paused: return

        key = Keyboard.as_event( key )

        if any( key in kb for kb in self.keyboards ):
            self._record_key( 'press', key )

        for kb in self.keyboards:
            kb.on_press( key )

    def on_release ( self, key : Union[str, KeyboardEvent] ):
        if self.paused: return

        key = Keyboard.as_event( key )

        if any( key in kb for kb in self.keyboards ):
            self._record_key( 'release', key )
        
        for kb in self.keyboards:
            kb.on_release( key )

    async def prompt_async ( self, prompt, cb = None ):
        self.paused = True

        try:
            res = await prompt()

            if callable( cb ): cb( res )

            return res
        finally:
            self.paused = False
        
    def prompt ( self, prompt, cb = None ):
        create_task( self.prompt_async( prompt, cb = cb ) )

    async def eval_async ( self, context : Context ):
        from .prompt import run_async

        await self.prompt_async( lambda: run_async( context, self.player ) )

    def eval ( self, context : Context ):
        create_task( self.eval_async( context ) )
        
    def add_source ( self, source : EventSource ):
        self.event_sources.append( source )

    async def listen ( self ):
        try:
            for source in self.event_sources:
                source.listen()

            await self.join_async()
        finally:
            for source in self.event_sources:
                source.close()

    def _serialize_event ( self, key : KeyboardEvent ):
        import json

        parameters = key.get_parameters()

        return ( key.__class__.__name__, key.serialize(), json.dumps( parameters, separators=(',', ':') ) if parameters else "" )

    def _deserialize_event ( self, class_name : str, event : str, parameters : str ) -> KeyboardEvent:
        import json

        if class_name in self.event_types:
            return self.event_types[ class_name ].deserialize( event, json.loads( parameters ) if parameters else {} )
        else:
            raise BaseException( f"Could not find event { class_name }" )

    def _record_key ( self, kind : str, key : KeyboardEvent ):
        file = self.record_file

        if file is not None:
            fd = self.record_fd

            if fd is None:
                fd = open( file, 'w' )

                self.assign_internal( 'record_fd', fd )

                self.assign_internal( 'record_start', self.player.get_time() )

                self.assign_internal( 'record_pressed', set() )
            
            pressed : Set[str] = self.lookup_internal( 'record_pressed' )

            key_str = str( key )

            if kind == 'press' and ( key_str not in pressed or not key.binary ):
                pressed.add( key_str )
            elif kind == 'release' and key_str in pressed:
                pressed.remove( key_str )
            else:
                return
            
            time = self.player.get_time() - self.record_start

            class_name, data, parameters = self._serialize_event( key )

            fd.write( f'{time},{kind},{class_name},{data},{parameters}\n' )

            fd.flush()

    def record ( self, file : str ):
        fd = self.record_fd

        if file is None and fd is not None:
            fd.close()

            self.assign_internal( 'record_fd', None )

        self.assign_internal( 'record_file', file )

    def readperf ( self, context : Context, file : str, keyboards : List[Keyboard] = None ) -> Music:
        with open( file, 'r' ) as f:
            entries = [ line.strip().split( ',' ) for line in list( f ) ]

        player = VirtualPlayer()

        for time, kind, class_name, data, parameters in entries:
            time = int( time )
            
            key : KeyboardEvent = self._deserialize_event( class_name, data, parameters )

            if kind == 'press':
                player.on_press( time, key )
            elif kind == 'release':
                player.on_release( time, key )
        
        if keyboards is None:
            return player.generate( context, self.keyboards )
        else:
            return player.generate( context, keyboards )

    def replay ( self, file : str, delay : int = 0 ):
        async def _replay ():
            with open( file, 'r' ) as f:
                entries = [ line.strip().split( ',' ) for line in list( f ) ]

            start = self.player.get_time()

            if delay > 0:
                await sleep( delay / 1000.0 )

            for time, kind, class_name, data, parameters in entries:
                time = int( time )
                
                key : KeyboardEvent = self._deserialize_event( class_name, data, parameters )

                await sleep( ( time - start ) / 1000.0 )

                start = time

                if kind == 'press':
                    self.on_press( key )
                elif kind == 'release':
                    self.on_release( key )

        create_task( _replay() )

    async def join_async ( self ):
        close_future : Future = self.lookup_internal( "close_future" )

        if close_future == None:
            close_future = Future()

            self.assign_internal( "close_future", close_future )

        await close_future



from .transformer import Transformer
from ..event import MusicEvent
from ..note import NoteEvent, NoteOnEvent, NoteOffEvent
from ..chord import ChordEvent, ChordOnEvent, ChordOffEvent
from ...music import MusicBuffer
from collections import defaultdict
from typing import Callable, Dict, List, Tuple, Any, Union, Optional, cast

class ComposeNotesTransformer( Transformer ):
    def __init__ ( self ):
        super().__init__()
        
        self.on_events : Dict[Any, Union[NoteOnEvent, ChordOnEvent]] = {}
        self.buffered_events : MusicBuffer = MusicBuffer()

        # Caches what is the earliest note_on we have received that is still waiting for the matching corresponding note_off
        # This allows us to quickly check if we can flush some of the events of the buffer
        # Without having to check every buffered on_events
        # Next Event Timestamp
        self.net : Optional[int] = None
        # Next Event Count
        self.nec : int = 0

    def _get_note_key ( self, event : Union[NoteOnEvent, NoteOffEvent] ):
        return ( event.voice.name, event.staff, int( event ) )

    def _off_event ( self, timestamp ):
        # If the next event timestamp is the same as the one we just composed
        # We need to update it and possibly flush the buffer
        if self.net == timestamp:
            if self.nec > 1:
                self.nec -= 1
            else:
                self.net = None
                self.nec = 0

                for event in self.on_events.values():
                    if self.nec == 0 or self.net > event.timestamp:
                        self.net, self.nec = ( event.timestamp, 1 )
                    elif self.net == event.timestamp:
                        self.nec = self.nec + 1

                for flushed_event in self.buffered_events.collect( None if self.nec == 0 else self.net ):
                    self.add_output( flushed_event )

    def _on_event ( self, timestamp : int ):
        if self.nec == 0 or self.net > timestamp:
            self.net, self.nec = ( timestamp, 1 )
        elif self.net == timestamp:
            self.nec = self.nec + 1

    def _on_note_off ( self, event : NoteOffEvent ):
        key = self._get_note_key( event )

        if key in self.on_events:
            on_event = cast( NoteOnEvent, self.on_events[ key ] )

            duration = event.timestamp - on_event.timestamp
            value = on_event.voice.from_duration_absolute( duration )

            composed_event = NoteEvent(
                timestamp = on_event.timestamp,
                pitch_class = on_event.pitch_class,
                value = value,
                duration = duration,
                octave = on_event.octave,
                voice = on_event.voice,
                velocity = on_event.velocity,
                accidental = on_event.accidental,
                tied = on_event.tied,
                staff = on_event.staff
            )

            self.buffered_events.append( composed_event )

            del self.on_events[ key ]

            self._off_event( composed_event.timestamp )
    
    def _on_note_on ( self, event : NoteOnEvent ):
        self.on_events[ self._get_note_key( event ) ] = event

        self._on_event( event.timestamp )

    def _get_chord_key ( self, event : Union[ChordOnEvent, ChordOffEvent] ):
        return ( event.voice.name, event.staff, tuple( event.pitches ) )

    def _on_chord_off ( self, event : ChordOffEvent ):
        key = self._get_chord_key( event )

        if key in self.on_events:
            on_event = cast( ChordOnEvent, self.on_events[ key ] )

            duration = event.timestamp - on_event.timestamp
            value = on_event.voice.from_duration_absolute( duration )

            composed_event = ChordEvent(
                timestamp = on_event.timestamp,
                pitches = on_event.pitches,
                name = on_event.name,
                value = value,
                duration = duration,
                voice = on_event.voice,
                velocity = on_event.velocity,
                tied = on_event.tied,
                staff = on_event.staff
            )

            self.buffered_events.append( composed_event )

            del self.on_events[ key ]

            self._off_event( composed_event.timestamp )
    
    def _on_chord_on ( self, event : ChordOnEvent ):
        self.on_events[ self._get_chord_key( event ) ] = event

        self._on_event( event.timestamp )

    def transform ( self ):
        while True:
            done, event = yield

            if done: break

            if isinstance( event, NoteOffEvent ):
                self._on_note_off( event )
            elif isinstance( event, NoteOnEvent ):
                self._on_note_on( event )
            elif isinstance( event, ChordOffEvent ):
                self._on_chord_off( event )
            elif isinstance( event, ChordOnEvent ):
                self._on_chord_on( event )
            elif self.buffered_events:
                self.buffered_events.append( event )
            else:
                self.add_output( event )

        for event in self.buffered_events.collect():
            self.add_output( event )


from musikla.core import Context, Instrument, Music
from musikla.parser.abstract_syntax_tree import MusicSequenceNode
from musikla.parser.abstract_syntax_tree.context_modifiers import ContextModifierNode
from musikla.graphics import BaseApplication
from musikla.libraries import KeyboardLibrary, KeyStroke, MusicLibrary, StandardLibrary
from musikla.parser import Parser
from musikla.audio import Player
from musikla.audio.sequencers import FluidSynthSequencer
from fractions import Fraction
import imgui
import glfw
import traceback
import time

EXPRESSION_TAB_AST = 0
EXPRESSION_TAB_EVENTS = 1
EXPRESSION_TAB_KEYBOARD = 2

class GuiApplication( BaseApplication ):
    def __init__ ( self ):
        with open( 'examples/minecraft.ml', 'r' ) as f:
            self.code = f.read()

        self.parsedTree = None
        self.parsedException = None
        self.player = None
        self.context = None

        self.expressionTab = EXPRESSION_TAB_AST

    def create_context ( self, player : Player ):
        ctx = Context.create()

        ctx.link( StandardLibrary() )
        ctx.link( MusicLibrary() )
        ctx.link( KeyboardLibrary( player ) )

        return ctx
    
    def imgui_tabbar ( self, open_tab, tabs ):
        width = imgui.get_content_region_available_width()

        tab_width = ( width - imgui.get_style().item_spacing.x * ( len( tabs ) - 1 ) ) / len( tabs )

        first = True

        open_panel = None

        for key, label, panel in tabs:
            if not first: 
                imgui.same_line()

            if imgui.button( label, width = tab_width ):
                open_tab = key

            if open_tab == key:
                open_panel = panel

            first = False

        if open_panel != None:
            open_panel()

        return open_tab
        
    def render_inspector_ast ( self ):
        imgui.begin_child( "inspector", 0, 0, border = True )
        if self.parsedTree != None:
            self.render_inspector( self.parsedTree )
        imgui.end_child()

    def render_inspector_events ( self ):
        imgui.begin_child( "notes", 0, 0, border = True )
        if self.player != None:
            for note in self.player.events:
                imgui.text( str( note ) )
        imgui.end_child()

    def get_pressed_keystrokes ( self ):
        for i in range( glfw.KEY_A, glfw.KEY_Z + 1 ):
            if imgui.get_io().keys_down[ i ]:
                yield KeyStroke( 
                    imgui.get_io().key_ctrl, 
                    imgui.get_io().key_alt, 
                    imgui.get_io().key_shift, 
                    chr( ord( 'a' ) + ( i - glfw.KEY_A ) )
                )

    def render_inspector_keyboard ( self ):
        imgui.begin_child( "keyboard", 0, 0, border = True )

        if self.context != None:
            current_keys = list( self.get_pressed_keystrokes() )

            keyboard : KeyboardLibrary = self.context.library( KeyboardLibrary )

            for key in current_keys:
                keyboard.on_press( key, self.player )

            released = set( keyboard.pressed_keys ) - set( current_keys )
            
            for key in released:
                keyboard.on_release( key, self.player )

            for key, expr in keyboard.keys:
                imgui.bullet()

                if expr.is_pressed: imgui.text_colored( str( key ), 0, 1, 0 )
                else: imgui.text( str( key ) )

        imgui.end_child()

    def render ( self ):
        super().render()

        if imgui.begin( "Music Editor" ):
            ( width, height ) = imgui.get_content_region_available();

            ( changed, value ) = imgui.input_text_multiline( "###Code", self.code, 1000, width = width, height = height / 2 )
            if changed: self.code = value
            
            to_parse = False
            to_play = False

            if imgui.button( "Parse" ):
                to_parse = True

            imgui.same_line()
            
            if imgui.button( "Play" ):
                to_play = True

            try:
                if to_parse or to_play:
                    self.player = Player()

                    self.player.sequencers.append( FluidSynthSequencer() )

                    parser = Parser();

                    self.parsedTree = parser.parse( self.code )
                    
                    self.context = self.create_context( self.player )

                    value = self.parsedTree.eval( self.context )
                    
                    self.player.events = events = list( value ) if isinstance( value, Music ) else []

                    self.parsedException = None
                if to_play:
                    self.player.play()
            except Exception as ex:
                self.parsedException = ex
                print(ex)
                traceback.print_tb(ex.__traceback__)

            if self.parsedException != None:
                imgui.text_colored( str( self.parsedException ), 1, 0, 0 )

            
            self.expressionTab = self.imgui_tabbar( self.expressionTab, [
                ( EXPRESSION_TAB_AST, "AST", self.render_inspector_ast ),
                ( EXPRESSION_TAB_EVENTS, "Events", self.render_inspector_events ),
                ( EXPRESSION_TAB_KEYBOARD, "Keyboard", self.render_inspector_keyboard )
            ] )

            imgui.end()

    def render_inspector ( self, obj, prefix = None ):
        properties = obj.__dict__ if hasattr( obj, '__dict__' ) else obj

        node_name = f"{obj.__class__.__name__}###{id( obj )}";

        if imgui.tree_node( node_name if prefix == None else f"{prefix}: {node_name}", imgui.TREE_NODE_DEFAULT_OPEN ):
            for key, value in properties.items():
                self.render_inspector_value( key, value );

            imgui.tree_pop()

    def render_inspector_value ( self, key, value ):
        if isinstance( value, int )\
        or isinstance( value, bool )\
        or isinstance( value, float )\
        or isinstance( value, str )\
        or isinstance( value, tuple )\
        or isinstance( value, Fraction ) \
        or value is None:
            imgui.bullet_text( f"{key}: {value}" )
        elif isinstance( value, list ):
            if imgui.tree_node( f"{ key }: { value.__class__.__name__ }({ len( value ) } items)###{ id( value ) }", imgui.TREE_NODE_DEFAULT_OPEN ):
                i = 0;

                for obj in value:
                    self.render_inspector_value( str( i ), obj );

                    i += 1

                imgui.tree_pop()
        else:
            self.render_inspector( value, key )



from musikla.core.music import MusicBuffer
from .transformer import Transformer

class SortTransformer(Transformer):
    def __init__ ( self, window_size : int = 0 ):
        self.buffer : MusicBuffer = MusicBuffer()
        self.window_size : int = window_size
        self.last_timestamp : int = 0

    def transform ( self ):
        while True:
            done, event = yield

            if done: break

            if self.window_size == 0:
                self.add_output( event )
            else:
                self.buffer.append( event )

                if event.timestamp > self.last_timestamp:
                    self.last_timestamp = event.timestamp

                    for ev in self.buffer.collect( self.last_timestamp - self.window_size ):
                        self.add_output( ev )
                    
        for ev in self.buffer.collect():
            self.add_output( ev )
        
        self.end_output()





from typing import Callable
from musikla.core.symbols_scope import SymbolsScope
from typing import Any, Dict, List, Optional
from musikla.core import Context, Music
from musikla.parser.abstract_syntax_tree import Node
from musikla.audio import PlayerLike, InteractivePlayer
from asyncio import create_task
from .event import KeyboardEvent

class KeyAction:
    def __init__ ( self, key : KeyboardEvent, expr : Node, args : List[str], context : Context, hold : bool = False, toggle : bool = False, repeat : bool = False, extend : bool = False, release : bool = False ):
        self.key : KeyboardEvent = key
        self.expr : Node = expr
        self.args : List[str] = args
        self.context : Context = context
        self.hold : bool = hold
        self.toggle : bool = toggle
        self.repeat : bool = repeat
        self.extend : bool = extend
        self.release : bool = release

        self.is_pressed : bool = False

        self.interactive_player : Optional[InteractivePlayer] = None
        self.sync : bool = False

    @property
    def is_playing ( self ):
        return self.interactive_player is not None and self.interactive_player.is_playing

    def play ( self, context : Context, player : PlayerLike, parameters : Dict[str, Any], cb : Callable = None ):
        forked_context : Optional[Context] = None

        forked_symbols : SymbolsScope = self.context.symbols.fork( opaque = False )

        for arg in self.args:
            forked_symbols.assign( arg, parameters[ arg ] if arg in parameters else None, local = True )

        def eval ():
            nonlocal forked_context, forked_symbols

            now = player.get_time() if forked_context == None else forked_context.cursor

            forked_context = self.context.fork( cursor = now, symbols = forked_symbols )

            value = forked_context.script.eval( self.expr, context = forked_context )

            if isinstance( value, Music ):
                return value.expand( forked_context )
            elif callable( value ):
                from musikla.core.callable_python_value import CallablePythonValue

                value = CallablePythonValue.call( value, forked_context )

                if isinstance( value, Music ):
                    return value.expand( forked_context )

            return None

        self.interactive_player = InteractivePlayer( eval, player, 0, self.repeat and not self.extend, self.extend, realtime = True )

        if cb is not None:
            cb( self.interactive_player )

        if self.sync:
            self.interactive_player.start_sync()
        else:
            create_task( self.interactive_player.start() )

    def stop ( self, context : Context, player : PlayerLike ):
        if self.interactive_player != None:
            if self.sync:
                self.interactive_player.stop_sync( player.get_time() )
            else:
                create_task( self.interactive_player.stop() )

        self.interactive_player = None

    def on_press ( self, context : Context, player : PlayerLike, parameters : Dict[str, Any] = {}, cb : Callable = None ):
        binary = self.key.binary

        if not binary and self.is_pressed:
            if self.hold or self.toggle:
                return self.on_release( context, player, cb )
            else:
                self.is_pressed = False

        if self.is_pressed and not ( binary and self.release ):
            return

        if not self.is_pressed and ( binary and self.release ):
            self.is_pressed = True
            return

        self.is_pressed = True
        
        if self.toggle:
            if self.interactive_player != None and self.interactive_player.is_playing:
                self.stop( context, player )
            else:
                self.play( context, player, parameters, cb )
        else:
            self.play( context, player, parameters, cb )
        
    def on_release ( self, context : Context, player : PlayerLike, cb : Callable = None ):
        if not self.is_pressed:
            return

        if self.key.binary and self.release:
            self.on_press( context, player, {}, cb )
            self.is_pressed = False
            return

        self.is_pressed = False

        if self.hold and not ( self.key.binary and self.release ):
            self.stop( context, player )

    def clone ( self, **kargs ) -> 'KeyAction':
        action = KeyAction(
            key = self.key,
            expr = self.expr,
            args = self.args,
            context = self.context,
            toggle = self.toggle,
            hold = self.hold,
            repeat = self.repeat,
            extend = self.extend,
            release = self.release
        )

        for key, value in kargs.items():
            setattr( action, key, value )
        
        return action


from musikla.parser.printer import CodePrinter
from musikla.core import Context, SymbolsScope, CallableValue, Value
from .statement_node import StatementNode
from ..expressions import VariableExpressionNode
from ..stack_frame_node import StackFrameNode
from typing import Tuple

class FunctionDeclarationStatementNode( StatementNode ):
    def __init__ ( self, name, arguments, body, using = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.name = name
        self.arguments = arguments
        self.using = using or []
        self.body = StackFrameNode( body, position = position )

    def __eval__ ( self, context : Context ):
        fn = CallableValue( lambda *args, **kargs: self.exec( context.symbols, *args, **kargs ) )

        if self.name != None:
            context.symbols.assign( self.name, fn )

        return fn

    def exec ( self, symbols_scope : SymbolsScope, context : Context, *args, **kargs ):
        forked = context.fork( symbols = symbols_scope.fork() )

        forked.symbols.assign( '__callerctx__', context, local = True )

        for variable_name in self.using:
            forked.symbols.using( variable_name )

        for i in range( len( self.arguments ) ):
            ( name, arg_mod, default_value ) = self.arguments[ i ]

            argument_node, default_node = None, default_value

            if len( args ) > i:
                argument_node = args[ i ]
            elif name in kargs:
                argument_node = kargs[ name ]
            elif default_value == None:
                raise Exception( f"Mandatory argument { name } was not given." )

            if arg_mod == 'expr':
                forked.symbols.assign( name, argument_node )
            elif ( arg_mod == 'ref' or arg_mod == 'in' ) and argument_node is not None:
                is_variable = isinstance( argument_node, VariableExpressionNode )

                if arg_mod == 'ref' and not is_variable:
                    raise BaseException( f"Only variable references can be passed to a function (function { self.name }, parameter { name })" )

                # TODO: Should ref pointers be shallow? Since ref can mutate
                # the value, it could be sensible for them to be so
                if is_variable:
                    pointer = context.symbols.pointer( argument_node.name )

                    if pointer is None:
                        context.symbols.assign( argument_node.name, None )
                    else:
                        forked.symbols.using( pointer, name )
                else:
                    forked.symbols.assign( name, Value.assignment( argument_node.eval( context.fork() ) ) )
            else:
                if argument_node is not None:
                    value = Value.assignment( argument_node.eval( context.fork() ) )
                else:
                    value = Value.assignment( default_node.eval( forked.fork() ) )
                
                value = forked.symbols.assign( name, value )

        return self.body.eval( forked )
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'fun ' )

        if self.name is not None:
            printer.add_token( self.name + ' ' )

        printer.begin_block( '(', ')' )

        for i in range( len( self.arguments ) ):
            if i > 0:
                printer.add_token( ', ' )

            name, mod, expr = self.arguments[ i ]

            if mod is not None:
                printer.add_token( mod + ' ' )
            
            printer.add_token( '$' + name )

            if expr is not None:
                printer.add_token( ' = ' )

                expr.to_source( printer )

        printer.end_block()

        printer.begin_block()

        self.body.to_source( printer )

        printer.end_block()



from musikla.core.value import CallableValue
from musikla.parser.printer import CodePrinter
from ..node import Node
from .expression_node import ExpressionNode
from musikla.core import Value, Music
from typing import Optional, Tuple, Union

class BinaryOperatorNode( ExpressionNode ):
    def __init__ ( self, left : Node, right : Node, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.left : Node = left
        self.right : Node = right

    def __eval__ ( self, context ):
        return None

class PlusBinaryOperatorNode(BinaryOperatorNode):
    def __eval__ ( self, context ):
        left_value = self.left.eval( context )
        right_value = self.right.eval( context )

        return left_value + right_value

    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' + ' )

        self.right.to_source( printer )

class MinusBinaryOperatorNode(BinaryOperatorNode):
    def __eval__ ( self, context ):
        left_value = self.left.eval( context )
        right_value = self.right.eval( context )

        return left_value - right_value
    
    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' - ' )

        self.right.to_source( printer )

class PowBinaryOperatorNode(BinaryOperatorNode):
    def __eval__ ( self, context ):
        left_value = self.left.eval( context.fork(cursor = 0) )
        right_value = self.right.eval( context.fork(cursor = 0) )

        return left_value ** right_value
    
    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' ** ' )

        self.right.to_source( printer )

class MultBinaryOperatorNode(BinaryOperatorNode):
    def __eval__ ( self, context ):
        left_value = self.left.eval( context.fork( cursor = 0 ) )
        right_value = self.right.eval( context.fork( cursor = 0 ) )

        return left_value * right_value
    
    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' * ' )

        self.right.to_source( printer )

class DivBinaryOperatorNode(BinaryOperatorNode):
    def __eval__ ( self, context ):
        left_value = self.left.eval( context )
        right_value = self.right.eval( context )

        return left_value / right_value
    
    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' / ' )

        self.right.to_source( printer )

class AndLogicOperatorNode(BinaryOperatorNode):
    def __eval__ ( self, context ):
        left_value = self.left.eval( context )

        if not left_value: return False

        right_value = self.right.eval( context )

        return bool( right_value )

    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' and ' )

        self.right.to_source( printer )

class OrLogicOperatorNode(BinaryOperatorNode):
    def __eval__ ( self, context ):
        left_value = self.left.eval( context )

        if left_value:
            return left_value

        right_value = self.right.eval( context )

        return right_value if right_value else False

    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' or ' )

        self.right.to_source( printer )

class ComparisonOperatorNode(BinaryOperatorNode):
    operator : Optional[str] = None

    def __init__ ( self, left : Node, right : Node, position : Tuple[int, int, int] = None ):
        super().__init__( left, right, position )

    def compare ( self, a, b ):
        pass

    def __eval__ ( self, context, assignment : bool = False ):
        left_value = self.left.eval( context )
        right_value = self.right.eval( context )

        return self.compare( left_value, right_value )

    def to_source ( self, printer : CodePrinter ):
        self.left.to_source( printer )

        printer.add_token( ' ' + self.operator + ' ' )

        self.right.to_source( printer )

class GreaterEqualComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = '>='

    def compare ( self, a, b ): return a >= b

class GreaterComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = '>'

    def compare ( self, a, b ): return a > b

class EqualComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = '=='

    def compare ( self, a, b ): return a == b

class NotEqualComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = '!='

    def compare ( self, a, b ): return a != b

class LesserEqualComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = '<='

    def compare ( self, a, b ): return a <= b

class LesserComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = '<'

    def compare ( self, a, b ): 
        return a < b

class IsComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = 'in'

    def compare ( self, a, b ): 
        if isinstance( a, CallableValue ): a = a.raw()
        
        if isinstance( b, CallableValue ): b = b.raw()

        return a is b or type(a) is b

class IsNotComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = 'notin'

    def compare ( self, a, b ): 
        if isinstance( a, CallableValue ): a = a.raw()
        
        if isinstance( b, CallableValue ): b = b.raw()

        return ( a is not b ) and ( type(a) is not b )

class InComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = 'in'

    def compare ( self, a, b ): 
        return a in b

class NotInComparisonOperatorNode(ComparisonOperatorNode):
    operator : str = 'notin'

    def compare ( self, a, b ): 
        return a not in b



import unittest
from musikla.parser.printer import CodePrinter
from musikla.parser.parser import Parser

parser = Parser()

class ParserTestCase(unittest.TestCase):
    def setUp ( self ):
        self.parser = parser
        self.printer = CodePrinter()

    def assertParse ( self, source, expected = None ):
        if expected is None:
            expected = source

        ast = self.parser.parse( source )

        source_printed = self.printer.print( ast )

        self.assertEqual( source_printed, expected )

    def test_number ( self ):
        self.assertParse( "1" )
        self.assertParse( "1.2" )
        self.assertParse( "+1", "1" )
        self.assertParse( "-1" )

    def test_string ( self ):
        self.assertParse( "'ola'" )
        self.assertParse( '"ola"', "'ola'" )

    def test_notes ( self ):
        self.assertParse( "a" )
        self.assertParse( "A" )
        self.assertParse( "C," )
        self.assertParse( "C,/2", "C,1/2" )
        self.assertParse( "C,2/2", "C," )
        self.assertParse( "C,3/2" )
        self.assertParse( "C,3" )
        self.assertParse( "c'''" )
        self.assertParse( "__c'''" )
        self.assertParse( "_c'''" )
        self.assertParse( "^^c'''" )
        self.assertParse( "^c'''" )

    def test_chords ( self ):
        self.assertParse( "[CFG,]" )
        
        # Shortcuts
        self.assertParse( "[Fm3]" )
        self.assertParse( "[FM3]" )            
        self.assertParse( "[Fm7]" )
        self.assertParse( "[FM7]" )
        self.assertParse( "[Fdom7]" )
        self.assertParse( "[F7]" )
        self.assertParse( "[Fm7b5]" )
        self.assertParse( "[Fdim7]" )
        self.assertParse( "[FmM7]" )
        self.assertParse( "[F5]" )
        self.assertParse( "[FM]" )
        self.assertParse( "[Fm]" )
        self.assertParse( "[Faug]" )
        self.assertParse( "[Fdim]" )
        self.assertParse( "[F]" )

        self.assertParse( "[F+]" )
        self.assertParse( "[F+]1/2" )
        self.assertParse( "[F__G^A]4/3" )

    def test_attribution ( self ):
        self.assertParse( "$a = 1" )
        self.assertParse( "$a::m = 1" )
        self.assertParse( "$a::m::[1] = 1" )
        self.assertParse( "$a::m::n::[1] = 1" )
        self.assertParse( "($a::m::n)::[1] = 1" )

    def test_if_statement ( self ):
        self.assertParse( "if (1) {2} else {3}" )
        self.assertParse( "if (1) 2 else 3" )
        self.assertParse( "if 1 {2} else 3" )
        self.assertParse( "if 1 then 2 else 3" )
        self.assertParse( "if 1 then 2 else if 3 then 4 else 5" )

    def test_while_statement ( self ):
        self.assertParse( "while (1) {2}" )
        self.assertParse( "while (1) 2" )
        self.assertParse( "while 1 {2}" )
        self.assertParse( "while 1 then 2" )
        



from musikla.parser.printer import CodePrinter
from typing import Tuple
from .expression_node import ExpressionNode
from ..node import Node
from ..stack_frame_node import StackFrameNode
from musikla.core import Context

class BlockNode( ExpressionNode ):
    def __init__ ( self, body : Node = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.body = StackFrameNode( body, position = position )
        self.fork_context : bool = True
        self.create_stack_frame : bool = True

    def __eval__ ( self, context : Context ):
        forked = context.fork( symbols = context.symbols.fork( opaque = False ) ) if self.fork_context else context

        try:
            if self.create_stack_frame:
                return self.body.eval( forked )
            else:
                return self.body.child.eval( forked )
        finally:
            context.join( forked )

    def to_source ( self, printer : CodePrinter ):
        with printer.block():
            self.body.to_source( printer )



from musikla.parser.printer import CodePrinter
from typing import Any, Optional, Tuple, List
from .statement_node import StatementNode
from ..node import Node
from musikla.core import Value, Context

class MultiVariableDeclarationStatementNode( StatementNode ):
    def __init__ ( self, left : List[Node], right : Node, operator : Optional[str] = None, local : bool = False, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.left : Any = left

        self.right : Any = right
        self.operator : Optional[str] = operator
        self.local : bool = local

        for op in self.left:
            if self.operator is None and not hasattr( op, 'assign' ):
                raise BaseException( f"Left hand side \"{CodePrinter().print(op)}\" cannot be used in an attribution" )
            elif self.operator is not None and not hasattr( op, 'lookup_assign' ):
                raise BaseException( f"Left hand side \"{CodePrinter().print(op)}\" cannot be used in an attribution" )

    def __eval__ ( self, context : Context ):
        val = Value.assignment( self.right.eval( context.fork( cursor = 0 ) ) )

        for i, op in enumerate( self.left ):
            if self.operator is None or self.operator == "":
                op.assign( context, val[ i ], local = self.local )
            else:
                def _set ( value ):
                    nonlocal val

                    if self.operator == '*': value *= val[ i ]
                    elif self.operator == '/': value /= val[ i ]
                    elif self.operator == '+': value += val[ i ]
                    elif self.operator == '-': value -= val[ i ]
                    elif self.operator == '&': value &= val[ i ]
                    elif self.operator == '|': value |= val[ i ]
                    else: raise Exception( "Invalid operator: " + self.operator )

                    return value

                op.lookup_assign( context, _set, local = self.local )

        return None

    def to_source ( self, printer : CodePrinter ):
        for i, op in enumerate( self.to_source ):
            if i > 0:
                printer.add_token( ", " )

            op.to_source( printer )

        printer.add_token( f" { self.operator if self.operator is not None else '' }= " )
        
        self.right.to_source( printer )



from typing import Tuple
from musikla.parser.printer import CodePrinter
from musikla.core import Context
from ..node import Node
from ..music_node import MusicSequenceBase

class WhileLoopStatementNode( MusicSequenceBase ):
    def __init__ ( self, condition : Node, body : Node, position : Tuple[int, int, int] = None ):
        from ..expressions.block_node import BlockNode

        super().__init__( position )

        self.condition : Node = condition
        self.body : Node = body

        if isinstance( self.body, BlockNode ):
            self.body.fork_context = False
            self.body.create_stack_frame = False

    def values ( self, context : Context ):
        condition_value = self.condition.eval( context )

        while condition_value:
            forked = context.fork( symbols = context.symbols.fork( opaque = False ) )

            yield self.body.eval( forked )

            context.join( forked )

            condition_value = self.condition.eval( context )

    def to_source ( self, printer : CodePrinter ):
        from ..expressions.group_node import GroupNode
        from ..expressions.block_node import BlockNode

        printer.add_token( 'while ' )

        self.condition.to_source( printer )

        if not isinstance( self.condition, GroupNode ) and not isinstance( self.body, BlockNode ):
            printer.add_token( " then " )
        else:
            printer.add_token( " " )

        self.body.to_source( printer )



from typing import Union, Optional
from fractions import Fraction
from .context import Context
from .music import Music, MusicBuffer
from .events import MusicEvent

class Metronome:
    def __init__ ( self, context : Context, beats : int = None, tick : Union[Music, MusicEvent] = None ):
        self.context : Context = context
        self.beats : Optional[int] = beats
        # self.duration : Optional[float] = float( duration ) if duration is not None else None
        self.tick : Optional[MusicEvent] = tick[ 0 ] if isinstance( tick, Music ) else tick
        
    def clone ( self ) -> 'Metronome':
        return Metronome( self.context, self.beats, self.tick )

    def _generate ( self, context : Context ):
        beats = self.beats or 0

        cursor = context.cursor

        beat_duration = self.context.voice.get_beat_duration()

        while self.beats is None or beats > 0:
            yield self.tick.clone( cursor = cursor )

            if beats > 0: beats -= 1

            cursor += beat_duration
    
    def to_music ( self, context : Context ) -> Music:
        return Music( self._generate( context ) )


from decimal import InvalidOperation
from musikla.core import Context, Value, Music
from musikla.core.theory import Note
from musikla.core.events import NoteEvent
from typing import Callable, List, Dict, Optional, Union, Any, cast
from musikla.parser.abstract_syntax_tree import Node, MusicSequenceNode
from musikla.parser.abstract_syntax_tree.expressions import FunctionExpressionNode, ConstantNode
from musikla.audio import Player, InteractivePlayer
from .event import KeyStroke, PianoKey, KeyboardEvent
from .action import KeyAction

class SubscriberEvent:
    def __init__ ( self ):
        self.listeners : List[Callable] = []

    def add_listener ( self, fn : Callable ):
        self.listeners.append( fn )

    def __iadd__ ( self, fn ):
        if callable( fn ):
            return self.add_listener( fn )
        
        return NotImplemented    

    def remove_listener ( self, fn : Callable ):
        if fn in self.listeners:
            self.listeners.remove( fn )
    
    def __isub__ ( self, fn ):
        if callable( fn ):
            return self.remove_listener( fn )
        
        return NotImplemented    

    def trigger ( self, *args ):
        for obs in self.listeners:
            obs( *args )

    def __call__ ( self, *args ):
        self.trigger( *args )

class KeyActionSubscriber:
    def __init__ ( self, 
            on_press : Callable = None,
            on_release : Callable = None,
            on_start : Callable = None,
            on_stop : Callable = None
        ):
        self.on_press : Optional[Callable] = on_press
        self.on_release : Optional[Callable] = on_release
        self.on_start : Optional[Callable] = on_start
        self.on_stop : Optional[Callable] = on_stop
    
class Keyboard:
    @staticmethod
    def as_event ( key_value : Any ) -> KeyboardEvent:
        if type( key_value ) == str:
            return KeyStroke.parse( key_value )
        elif isinstance( key_value, Music ) or isinstance( key_value, NoteEvent )  or isinstance( key_value, Note ):
            return PianoKey( key_value )
        elif isinstance( key_value, KeyboardEvent ):
            return key_value
        else:
            raise Exception( "Keyboard value is invalid" )

    def __init__ ( self, context : Context ):
        self.context : Context = context        
        self._player : Optional[Player] = None

        self.keys : Dict[KeyboardEvent, KeyAction] = dict()
        self.global_flags : Dict[str, int] = dict()
        self.global_prefixes : List[Node] = list()

        self._on_new_player_observers : List[Callable] = []

        self.manual_lifetime : bool = False

    def on_new_player ( self, player : InteractivePlayer ):
        for obs in self._on_new_player_observers:
            obs( self, player )

    def add_new_player_observer ( self, observer : Callable ):
        self._on_new_player_observers.append( observer )
    
    def remove_new_player_observer ( self, observer : Callable ):
        if observer in self._on_new_player_observers:
            self._on_new_player_observers.remove( observer )

    @property
    def player ( self ) -> Player:
        if self._player is None:
            from .library import KeyboardLibrary

            lib : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

            self._player = lib.player

        return self._player

    def get_keyboard_flag ( self, context : Context, node : Optional[Node], name : str ) -> bool:
        if node != None:
            value : Value = node.eval( context )

            return bool( value )

        return name in self.global_flags and self.global_flags[ name ] > 0

    def register_key ( self, context : Context, key : Node, expression : Node, args : List[str] = [], toggle : Node = None, hold : Node = None, repeat : Node = None, extend : Node = None, release : Node = None ):
        toggle_value = self.get_keyboard_flag( context, toggle, "toggle" )
        hold_value = self.get_keyboard_flag( context, hold, "hold" )
        repeat_value = self.get_keyboard_flag( context, repeat, "repeat" )
        extend_value = self.get_keyboard_flag( context, extend, "extend" )
        release_value = self.get_keyboard_flag( context, release, "release" )
        
        key_value = key.eval( context )

        if self.global_prefixes:
            expression = MusicSequenceNode( [ *self.global_prefixes, expression ] )

        key_event = Keyboard.as_event( key_value )

        action = KeyAction(
            key = key_event,
            expr = expression,
            args = args,
            context = context,
            toggle = toggle_value,
            hold = hold_value,
            repeat = repeat_value,
            extend = extend_value,
            release = release_value
        )

        self.keys[ action.key ] = action

    def push_flags ( self, context : Context, *flags : Node ):
        for flag in flags:
            value : str = flag.eval( context )

            if Value.typeof( value ) == str:
                if value in self.global_flags:
                    self.global_flags[ value ] += 1
                else:
                    self.global_flags[ value ] = 1

    def pop_flags ( self, context : Context, *flags : Node ):
        for flag in flags:
            value : str = flag.eval( context )

            if Value.typeof( value ) == str:
                if value in self.global_flags:
                    self.global_flags[ value ] -= 1

                    if self.global_flags[ value ] == 0:
                        del self.global_flags[ value ]

    def push_prefix ( self, context : Context, expression : Node ):
        self.global_prefixes.append( expression )

    def pop_prefix ( self, context : Context ):
        self.global_prefixes.pop()

    def start_all ( self ):
        for key in self.keys.values():
            key.play( self.context, self.player, {} )

    def stop_all ( self ):
        for key in self.keys.values():
            key.stop( self.context, self.player )

    def start ( self, key : Union[ KeyboardEvent, str ] ):
        key_stroke : KeyboardEvent = cast( KeyboardEvent, KeyStroke.parse( key ) if Value.typeof( key ) == str else key )

        if key_stroke in self.keys:
            self.keys[ key_stroke ].play( self.context, self.player, key_stroke.get_parameters(), self.on_new_player )

    def stop ( self, key : Union[ KeyboardEvent, str ] ):
        key_stroke : KeyboardEvent = cast( KeyboardEvent, KeyStroke.parse( key ) if Value.typeof( key ) == str else key )

        if key_stroke in self.keys:
            self.keys[ key_stroke ].stop( self.context, self.player )

    def __getitem__ ( self, key ):
        key_stroke : KeyboardEvent = cast( KeyboardEvent, KeyStroke.parse( key ) if Value.typeof( key ) == str else key )

        return self.keys[ key_stroke ]

    def __contains__ ( self, key : Union[ KeyboardEvent, str ] ) -> bool:
        key_stroke : KeyboardEvent = cast( KeyboardEvent, KeyStroke.parse( key ) if Value.typeof( key ) == str else key )

        return key_stroke in self.keys

    def on_press ( self, key : Union[ KeyboardEvent, str ] ):
        key_stroke : KeyboardEvent = cast( KeyboardEvent, KeyStroke.parse( key ) if Value.typeof( key ) == str else key )

        if key_stroke in self.keys:
            self.keys[ key_stroke ].on_press( self.context, self.player, key_stroke.get_parameters(), self.on_new_player )

    def on_release ( self, key : Union[str, KeyboardEvent] ):
        key_stroke : KeyboardEvent = cast( KeyboardEvent, KeyStroke.parse( key ) if Value.typeof( key ) == str else key )

        if key_stroke in self.keys:
            self.keys[ key_stroke ].on_release( self.context, self.player, self.on_new_player )

    @property
    def is_closed ( self ) -> bool:
        from .library import KeyboardLibrary

        lib : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )
        
        return self not in lib.keyboards

    def close ( self, closing : bool = False ) -> 'Keyboard':
        from .library import KeyboardLibrary

        self.stop_all()

        if not closing:
            keyboard : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

            keyboard.close( self )
        
        return self

    def open ( self ) -> 'Keyboard':
        from .library import KeyboardLibrary

        lib : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )
        
        if self not in lib.keyboards:
            lib.keyboards.append( self )

        return self

    def _assert_keyboard ( self, obj ):
        if obj is None:
            raise InvalidOperation( "Cannot combine a keyboard with 'None'" )
            
        if not isinstance( obj, Keyboard ):
            raise InvalidOperation( f"Cannot combine a keyboard with '{ type( obj ) }'" )

    def clone ( self, auto_close : bool = True ):
        from .library import KeyboardLibrary

        lib : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

        keyboard = lib.create()
        
        if auto_close and not self.manual_lifetime:
            self.close()

        keyboard.context = self.context
        keyboard.keys = dict( self.keys )
        keyboard.global_flags = dict( self.global_flags )
        keyboard.global_prefixes = list( self.global_prefixes )
        keyboard.manual_lifetime = self.manual_lifetime
        
        return keyboard

    def map_actions ( self, context : Context, mapper : Node ) -> 'Keyboard':
        mapper = Value.eval( context, mapper ).raw() if isinstance( mapper, Node ) else mapper

        keyboard = self.clone()

        for key, action in self.keys.items():
            keyboard.keys[ key ] = mapper( context, ConstantNode( action ) )

        return keyboard

    def map ( self, context : Context, mapper : Node ) -> 'Keyboard':
        mapper = Value.eval( context, mapper ) if isinstance( mapper, Node ) else mapper

        keyboard = self.clone()

        for key, action in self.keys.items():
            keyboard.keys[ key ] = action.clone(
                expr = FunctionExpressionNode( ConstantNode( mapper ), [ ConstantNode( action.key ), action.expr ] )
            )

        return keyboard

    def with_grid ( self, context : Context, grid, mode : str = 'start_end' ) -> 'Keyboard':
        def _mapper ( context : Context, k, music ):
            return grid.align( context, music, mode )

        return self.map( context, _mapper )

    def __add__ ( self, other ):
        from .library import KeyboardLibrary

        self._assert_keyboard( other )

        keyboard : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

        kb = keyboard.create()

        kb += self
        kb += other

        return kb

    def __sub__ ( self, other ):
        from .library import KeyboardLibrary

        self._assert_keyboard( other )

        keyboard : KeyboardLibrary = cast( KeyboardLibrary, self.context.library( KeyboardLibrary ) )

        kb = keyboard.create()

        kb += self
        kb -= other

        return kb

    def __iadd__ ( self, other ):
        self._assert_keyboard( other )

        if not other.manual_lifetime:
            other.close()

        self.keys.update( other.keys )

        return self
    
    def __isub__ ( self, other ):
        self._assert_keyboard( other )

        for key in other.keys:
            self.keys.pop( key, None )

        return self

# class ActionMapNode(Node):
#     def __init__ ( self, mapper, key : KeyboardEvent, original : Node ):
#         self.mapper = mapper
#         self.key : KeyboardEvent = key
#         self.original : Node = original
    
#     def eval ( self, context : Context ):
#         result = Value.eval( context, self.original )

        

#         return self.mapper( self.key, result )



from musikla.parser.printer import CodePrinter
from typing import Tuple
from .context_modifier_node import ContextModifierNode
from musikla.core.events import ContextChangeEvent
from musikla.core import Context, Voice, Instrument

class InstrumentModifierNode( ContextModifierNode ):
    def __init__ ( self, instrument, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.instrument : int = instrument

    def apply ( self, voice : Voice ):
        voice.instrument = Instrument.from_program( self.instrument )

    def modify ( self, context : Context ):
        context.voice = context.voice.clone( instrument = Instrument.from_program( self.instrument ) )

        yield ContextChangeEvent( context.cursor, "instrument", context.voice.instrument.program, context.voice, 0 )

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'I' + str( self.instrument ) )



from musikla.parser.printer import CodePrinter
from .expression_node import ExpressionNode
from typing import Any, Tuple

class ConstantNode( ExpressionNode ):
    def __init__ ( self, value : Any, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.value : Any = value

    def __eval__ ( self, context ):
        return self.value

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( str( self.value ) )



from asyncio.tasks import sleep
import sys
import argparse
import os

from musikla.core import Context
from musikla.libraries import KeyboardLibrary, MidiLibrary
from musikla.audio import Player
from musikla.audio.sequencers import Sequencer, SequencerFactory
from musikla.parser.abstract_syntax_tree.expressions import Object
from musikla import Script
from typing import List, Optional, cast
from colorama import init
from shlex import shlex, split

class CliApplication:
    default_output = 'pulseaudio' if os.name == 'posix' else 'dsound'

    def __init__ ( self, argv ):
        self.argv = argv

    def enable_echo(self, fd, enabled):
        if os.name == 'posix':
            import termios
            
            (iflag, oflag, cflag, lflag, ispeed, ospeed, cc) \
                = termios.tcgetattr(fd)

            if enabled:
                lflag |= termios.ECHO
            else:
                lflag &= ~termios.ECHO

            new_attr = [iflag, oflag, cflag, lflag, ispeed, ospeed, cc]
            termios.tcsetattr(fd, termios.TCSANOW, new_attr)

    async def keyboard ( self, context : Context, virtual_keyboard : KeyboardLibrary ):
        print( "Keyboard active." )

        try:
            self.enable_echo( sys.stdin.fileno(), False )

            await virtual_keyboard.listen()
        finally:
            self.enable_echo( sys.stdin.fileno(), True )

    def split_argv ( self, argv : List[str], separators : List[str] ) -> List[List[str]]:
        groups : List[List[str]] = []

        start = 0

        for i in range( len( argv ) ):
            if argv[ i ] in separators:
                groups.append( argv[ start : i ] )

                start = i
        
        groups.append( argv[ start: ] )

        return groups

    def parse_outputs_args ( self, player : Player, argvs : List[List[str]] ) -> List[Sequencer]:
        sequencers : List[Sequencer] = []

        for argv in argvs:
            format : Optional[str] = None

            output : str = argv[ 1 ]
            
            if len( argv ) >= 4 and ( argv[ 2 ] == '-f' or argv[ 2 ] == '--format' ):
                format = argv[ 3 ]

                argv = argv[ 4: ]
            else:
                argv = argv[ 2: ]

            if format is not None:
                sequencers.append( player.make_sequencer_from_format( format, output, argv ) )
            else:
                sequencers.append( player.make_sequencer_from_uri( output, argv ) )

        return sequencers

    async def run ( self ):
        init()

        parser = argparse.ArgumentParser( description = 'Evaluate musical expression' )

        parser.add_argument( 'file', type = str, nargs = '?', help = 'Files to evaluate. No file means the input will be read from the stdin' )
        parser.add_argument( '-c', '--code', type = str, help = 'Execute a piece of code' )
        parser.add_argument( '-i', '--import', dest = 'imports', action = 'append', type = str, help = 'Import an additional library. These can be builtin libraries, or path to .ml and .py files' )
        parser.add_argument( '-o', '--output', dest = 'outputs', type = str, action = 'append', help = 'Where to output to. By default outputs the sounds to the device\'s speakers.' )
        parser.add_argument( '-f', '--format', dest = 'formats', type = str, action = 'append', help = 'Type of output to use' )
        parser.add_argument( '-v', '--variable', dest = 'variables', nargs=2, metavar=('name', 'value'), type = str, action = 'append', help = 'Set a custom variable available under $sys\\vars' )
        parser.add_argument( '--midi', type = str, help = 'Use a custom MIDI port by default when no name is specified' )
        parser.add_argument( '--soundfont', type = str, help = 'Use a custom soundfont .sf2 file' )
        parser.add_argument( '--print-events', dest = 'print_events', action='store_true', help = 'Print events (notes) to the console as they are played.' )
        parser.add_argument( '--profile', dest = 'profile', action='store_true', help = 'Measure and display total parse times' )
        parser.add_argument( '--traceback', dest = 'traceback', action='store_true', help = 'Display error tracebacks (aka stacktraces)' )
        parser.add_argument( '--skip-parser-cache', dest = 'skip_parser_cache', action='store_true', help = 'Force the parser to rebuild the grammar cache on disk' )

        if '--' in self.argv:
            index = self.argv.index( '--' )

            argv = self.argv[:index]
            internal_args = self.argv[index + 1:]
        else:
            argv = self.argv
            internal_args = []

        argv = self.split_argv( argv, [ '-o', '--output' ] )

        options = parser.parse_args( argv[ 0 ] )

        script = Script( symbols = {
            'sys\\args': internal_args,
            'sys\\vars': Object( options.variables or [] )
        }, parser_read_cache = not options.skip_parser_cache )

        if options.profile:
            print( "PARSE (AUTOLOAD): ", script.parser.time_spent * 1000, "ms" )

        if options.skip_parser_cache:
            script.parser.read_cache = False

        if options.traceback:
            script.print_tracebacks = True

        if argv[ 1: ]:
            sequencers = self.parse_outputs_args( script.player, argv[ 1: ] )
        elif script.config.has_option( 'Musikla', 'output' ):
            output_array = split( script.config.get( 'Musikla', 'output' ) )
            
            split_output_array = self.split_argv( output_array, [ '-o', '--output' ] )

            sequencers = self.parse_outputs_args( script.player, split_output_array[ 1: ] )
        else:
            sequencers = []

        script.player.print_events = bool( options.print_events )
        
        if not script.config.has_section( 'Musikla' ):
            script.config.add_section( 'Musikla' )

        if options.soundfont != None:
            script.config.set( 'Musikla', 'soundfont', options.soundfont )

        if options.midi != None:
            script.config.set( 'Musikla', 'midi_input', options.midi )

        if sequencers:
            for sequencer in sequencers:
                script.player.add_sequencer( sequencer )
        else:
            script.player.add_sequencer( self.default_output )

        for lib in options.imports or []:
            script.import_library( lib )

        if script.config.has_option( 'Musikla', 'midi_input' ):
            lib = cast( MidiLibrary, script.context.library( MidiLibrary ) )
            
            lib.set_midi_default_input( script.config.get( 'Musikla', 'midi_input' ) )

        try:
            if options.code is not None:
                script.execute( options.code, sync = False, realtime = script.player.realtime )
            elif options.file is None:
                # Super duper naive repl that accepts only a sequence of one-liners
                print( "Welcome to the Musikla REPL. Type \"quit();\" to exit." )
                print( '>>> ', end = '' )
                for line in sys.stdin:
                    try:
                        if line == 'quit();':
                            break

                        script.execute( line, sync = False, realtime = script.player.realtime )
                    except:
                        print( sys.exc_info()[ 0 ] )

                    print( '>>> ', end = '' )
            else:
                script.execute_file( options.file, sync = False, realtime = script.player.realtime )

            if options.profile:
                print( "PARSE (TOTAL): ", script.parser.time_spent * 1000, "ms" )

            keyboard : KeyboardLibrary = cast( KeyboardLibrary, script.context.library( KeyboardLibrary ) )

            # Wait for the end of the player if there is anything left to play
            await script.join()

            if keyboard != None and keyboard.has_keys:
                await self.keyboard( script.context, keyboard )
            else:
                await sleep(2)
        finally:
            script.player.close()



import unittest
from musikla.core import Context, Value
from musikla.parser.parser import Parser
from musikla.libraries.std import StandardLibrary

parser = Parser()

class EvalTestCase(unittest.TestCase):
    def setUp ( self ):
        self.parser = parser
        self.context = Context.create()
        self.context.link( StandardLibrary( None ) )

    def assertEval ( self, source, value ):
        ast = self.parser.parse( source )

        result = Value.eval( self.context, ast )

        self.assertEqual( result, value )

    def test_number ( self ):
        self.assertEval( "1", 1 )
        self.assertEval( "1.2", 1.2 )
        self.assertEval( "+1", 1 )
        self.assertEval( "-1", -1 )

    def test_string ( self ):
        self.assertEval( "'ola'", "ola" )
        self.assertEval( '"ola"', 'ola' )

    def test_if ( self ):
        self.assertEval( "if true then 1 else 2", 1 )
        self.assertEval( "if false then 1 else 2", 2 )
        self.assertEval( "if true then 1 else if true then 2 else 3", 1 )
        self.assertEval( "if false then 1 else if true then 2 else 3", 2 )
        self.assertEval( "if false then 1 else if false then 2 else 3", 3 )
    
    def test_operators ( self ):
        self.assertEval( "1 + 2", 3 )
        self.assertEval( "1 - 2", -1 )
        self.assertEval( "3 * 2", 6 )
        self.assertEval( "6 / 2", 3 )
        self.assertEval( "1 / 2", 0.5 )
        # Operator Precedence
        self.assertEval( "1 + 2 * 2", 5 )
        self.assertEval( "1 + 2 * 2 / 2", 3 )
        self.assertEval( "1 + 2 * 2 / 2 + 1", 4 )
        self.assertEval( "2 ** 3", 2 ** 3 )

        # Comparison Operators
        self.assertEval( "1 > 0", True )
        self.assertEval( "1 >= 0", True )
        self.assertEval( "1 < 0", False )
        self.assertEval( "1 <= 0", False )
        self.assertEval( "1 != 0", True )
        self.assertEval( "1 == 0", False )
        self.assertEval( "2 != 2", False )
        self.assertEval( "2 == 2", True )

        # Logical Operators
        self.assertEval( "true or false", True )
        self.assertEval( "false or false", False )
        self.assertEval( "true and false", False )
        self.assertEval( "true and true", True )
        self.assertEval( "not true or true", True )
        self.assertEval( "not true or false", False )

        self.assertEval( "true is $bool", True )
        self.assertEval( "true isnot $bool", False )
        self.assertEval( "none isnot none", False )
        self.assertEval( "0 is none", False )
        self.assertEval( "0 isnot none", True )

        self.assertEval( "0 in @[ 1, 2, 3 ]", False )
        self.assertEval( "0 notin @[ 1, 2, 3 ]", True )
        self.assertEval( "1 in @[ 1, 2, 3 ]", True )
        self.assertEval( "1 notin @[ 1, 2, 3 ]", False )

    def test_functions ( self ):
        self.assertEval( "fun fn ($a, $b = 2) => $a * $b; none", None )
        self.assertEval( "fn(4)", 8 )
        self.assertEval( "fn(4, 3)", 12 )
        self.assertEval( "fn(4, b = 3)", 12 )
        self.assertEval( "fn(a = 4, b = 3)", 12 )

        # Ref Parameters
        self.assertEval( "fun fn ( ref $a, $b ) { $a = $b }; $c = 2", None )
        self.assertEval( "fn($c; 4); $c", 4 )

    def test_assignment ( self ):
        self.assertEval( "$a = 1234; $a", 1234 )
        self.assertEval( "$a = 1234; { $a = 12 }; $a", 12 )
        self.assertEval( "$a = 10; while $a > 0 { $a -= 1 }; $a", 0 )
        self.assertEval( "$a = 0; while $a < 10 { $a += 1 }; $a", 10 )


from musikla.core.symbols_scope import Pointer
from musikla.parser.printer import CodePrinter
from musikla.core.events.event import MusicEvent
from musikla.core import Context, Library, CallableValue, Voice, Metronome, Instrument, Music, Value, Ref
from musikla.core.callable_python_value import CallablePythonValue
from musikla.core.events import ControlChangeEvent
from musikla.core.theory import Interval, Scale
from musikla.parser import Parser
from musikla.parser.abstract_syntax_tree import Node, MusicSequenceNode
from musikla.parser.abstract_syntax_tree.expressions import VariableExpressionNode
from musikla.parser.abstract_syntax_tree.context_modifiers import ContextModifierNode
from musikla.audio.sequencers.sequencer import Sequencer
from musikla.audio import Player
from typing import Any, List, Union, cast
from pathlib import Path
from fractions import Fraction
import musikla.audio.sequencers as seqs

def function_getctx ( context : Context ):
    return context

def function_withctx( _ : Context, ctx : Context, expr : Node ):
    return Value.eval( ctx, expr )

def function_pack ( context : Context, into = None, ctx = None, prefix : str = None, exclude = None, ignore_existing = False ):
    ctx = ctx or context
    into = into or type('PackedObject', (object,), {})()

    for key, value in ctx.symbols.enumerate( local = True ):
        if not key.startswith( '_' ) and ( exclude is None or key not in exclude ):
            p_key = ( prefix or "" ) + key

            if hasattr( into, p_key ) and not ignore_existing:
                raise BaseException( f"pack(): Cannot pack attribute '{ p_key }' of '{ type(into) }', attribute already exists." )

            if isinstance( value, Pointer ):
                value = value.scope.lookup( value.name )

            setattr( into, p_key, value )
    
    return into

def function_using ( context : Context, var ):
    if not isinstance( var, VariableExpressionNode ):
        raise BaseException( f"Using expected a variable syntax node" )

    context.symbols.using( var.name )

def function_play ( context : Context, expr ):
    value = expr.eval( context )

    if isinstance( value, Music ):
        return value

    return Music()

def function_discard ( context : Context, *expr ):
    for e in expr:
        value = e.eval( context.fork() )

        if isinstance(value, Music):
            for _ in value: pass

    return None

def function_ast ( context : Context, expr ):
    return expr

def function_ast_to_code ( ast : Node, ident = 4 ) -> str:
    printer = CodePrinter( ident = ident )

    return printer.print( ast )

def function_parse ( context : Context, code : str, rule : str = None ) -> Node:
    return context.script.parse( code, rule = rule )

def function_eval ( context : Context, code, rule : str = None ) -> Any:
    if type( code ) is str:
        code = function_parse( context, code, rule = rule )
    
    return Value.assignment( Value.eval( context, code ) )

def function_pyeval ( context : Context, code ) -> Any:
    from musikla.parser.abstract_syntax_tree.python_node import PythonNode
    
    return PythonNode.eval( context, code, True )

def function_pyexec ( context : Context, code ) -> Any:
    from musikla.parser.abstract_syntax_tree.python_node import PythonNode
    
    return PythonNode.execute( context, code, False )

SequencerLike = Union[str, List[str], Sequencer]

def function_make_sequencer ( context : Context, sequencer : SequencerLike ) -> Sequencer:
    std = cast( StandardLibrary, context.library( StandardLibrary ) )

    format = None
    uri = None
    args = None

    if type(sequencer) is str:
        uri = sequencer
    elif type( sequencer ) is list:
        args_start = 0

        if len( sequencer ) >= args_start + 2 and sequencer[ args_start ] == '-o':
            uri = sequencer[ args_start + 1 ]
            args_start += 2

        if len( sequencer ) >= args_start + 2 and sequencer[ args_start ] == '-f':
            format = sequencer[ args_start + 1 ]
            args_start += 2

        args = sequencer[ args_start: ]
    elif isinstance( sequencer, Sequencer ):
        return sequencer

    if format is not None:
        return std.player.make_sequencer_from_format( format, uri or "", args or [] )
    
    return std.player.make_sequencer_from_uri( uri or "", args or [] )

def function_save ( context : Context, music : Music, outputs : Union[str, Sequencer, List[SequencerLike]] ) -> Any:
    # The single parameter specifies if only one sequencer was given
    # In that case, instead of returning an array, we return just the sequencer
    single : bool = False

    if type( outputs ) is str or isinstance( outputs, Sequencer ):
        sequencers : List[Sequencer] = [ function_make_sequencer( context, outputs ) ]

        single = True
    else:
        sequencers : List[Sequencer] = [ function_make_sequencer( context, seq ) for seq in outputs ]

    for seq in sequencers: 
        seq.realtime = False

        seq.start()

    for ev in music.expand( context.fork( cursor = 0 ) ):
        for seq in sequencers:
            seq.register_event( ev )

    for seq in sequencers: seq.close()

    if single:
        return sequencers[ 0 ]

    return sequencers

def function_getcwd ():
    import os

    return os.getcwd()

def function_bool ( val ): return bool( val )

def function_int ( val ): return int( val )

def function_float ( val ): return float( val )

def function_str ( val ): return str( val )

def function_ord ( val ): return ord( val )

def function_chr ( val ): return chr( val )

def function_setvar ( var : Ref, value ):
    var.set( Value.assignment( value ) )

def function_hasattr ( o, name : str ):
    return hasattr( o, name )

def function_getattr ( o, name : str, default : Any = None ):
    return getattr( o, name, default = default )

def function_setattr ( o, name : str, value : Any ):
    return setattr( o, name, value )

def function_cc ( context : Context, control : int, value : int ):
    event = ControlChangeEvent( context.cursor, context.voice, control, value )
    
    return Music( [ event ] )

def function_gettime ( context : Context ):
    return context.cursor

def function_settime ( context : Context, time : int ):
    context.cursor = time

def function_slice ( notes : Music, start : int, end : int ):
    return notes.filter( lambda n: n.timestamp >= start and n.timestamp <= end )

def function_setvoice ( context : Context, voice : Voice ):
    context.voice = voice

def function_setinstrument ( context : Context, instrument : int, bank : int = None, soundfont : Union[int, str] = None ):
    context.voice.instrument = Instrument.from_program( instrument, bank, soundfont )

def function_sfload ( context : Context, soundfont : str, alias : str = None, only_new : bool = False ):
    script = context.script

    if not only_new:
        for seq in script.player.sequencers:
            if hasattr( seq, 'load_soundfont' ):
                cast( Any, seq ).load_soundfont( soundfont, alias )
    
    for factory in script.player.sequencer_factories:
        if hasattr( factory, 'add_soundfont' ):
            cast( Any, factory ).add_soundfont( soundfont, alias )

def function_sfunload ( context : Context, soundfont : str, only_new : bool = False ):
    script = context.script

    if not only_new:
        for seq in script.player.sequencers:
            if hasattr( seq, 'unload_soundfont' ):
                cast( Any, seq ).unload_soundfont( soundfont )
    
    for factory in script.player.sequencer_factories:
        if hasattr( factory, 'remove_soundfont' ):
            cast( Any, factory ).remove_soundfont( soundfont )

def function_debug ( context : Context, expr ):
    value = expr.eval( context.fork() )

    if value is None:
        print( None )
    elif isinstance( value, Music ):
        print( '<Music> ' + ' '.join( str( event ) for event in value.expand( context ) ) )
    else:
        print( "<%s>%s" % ( Value.typeof( value ), value ) )

def function_mod ( n : float, d : float ) -> float: 
    return n % d

def function_div ( n : float, d : float ) -> float: 
    return n // d

def function_inspect_context ( context : Context, ignore_root : bool = True, ignore_symbols : bool = False ):
    symbols = context.symbols
    
    while symbols != None and ( not ignore_root or symbols.parent is not None ):
        print( f"Context#{ id( symbols ) } (Opaque = { symbols.opaque })" )

        if not ignore_symbols:
            for container_name, container in symbols.symbols.items():
                print( f"  - { container_name or '<default>' }:" )
                for key, value in container.items():
                    print( f"    - { key }: { value }" )

        symbols = symbols.parent
    print()

def function_voices_create ( context : Context, name : str, modifiers : Node = None, inherit : Voice = None ):
    if inherit != None:
        pass

    voice = Voice( name, Instrument( name, 1 ) )
    forked = context.fork()
    forked.voice = voice

    if inherit != None:
        voice.instrument = inherit.instrument
        voice.time_signature = inherit.time_signature
        voice.velocity = inherit.velocity
        voice.octave = inherit.octave
        voice.value = inherit.value
        voice.tempo = inherit.tempo
    
    if modifiers != None:
        if isinstance( modifiers, MusicSequenceNode ):
            for modifier in modifiers:
                if isinstance( modifier, ContextModifierNode ):
                    modifier.apply( voice )
                else:
                    modifier.eval( forked )
        else:
            if isinstance( modifiers, ContextModifierNode ):
                modifiers.apply( voice )
            else:
                modifiers.eval( forked )

    return voice

def function_seek ( context : Context, time : Node ):
    context.cursor += Value.eval( context, time )

def function_len ( context : Context, obj : Any ):
    if isinstance( obj, Music ):
        return obj.len( context )
    else:
        return len( obj )

class StandardLibrary(Library):
    def __init__ ( self, player : Player ):
        super().__init__()

        self.player : Player = player
    
    def on_link ( self, script ):
        context : Context = self.context

        context.symbols.assign( "getctx", CallablePythonValue( function_getctx ) )
        context.symbols.assign( "withctx", CallablePythonValue( function_withctx ) )
        context.symbols.assign( "pack", CallablePythonValue( function_pack ) )
        context.symbols.assign( "print", CallablePythonValue( print ) )
        context.symbols.assign( "debug", CallableValue( function_debug ) )
        context.symbols.assign( "discard", CallableValue( function_discard ) )
        context.symbols.assign( "play", CallableValue( function_play ) )
        context.symbols.assign( "using", CallableValue( function_using ) )
        context.symbols.assign( "seek", CallableValue( function_seek ) )
        context.symbols.assign( "ast", CallableValue( function_ast ) )
        context.symbols.assign( "ast_to_code", CallablePythonValue( function_ast_to_code ) )
        context.symbols.assign( "parse", CallablePythonValue( function_parse ) )
        context.symbols.assign( "eval", CallablePythonValue( function_eval ) )
        context.symbols.assign( "pyeval", CallablePythonValue( function_pyeval ) )
        context.symbols.assign( "pyexec", CallablePythonValue( function_pyexec ) )
        context.symbols.assign( "make_sequencer", CallablePythonValue( function_make_sequencer ) )
        context.symbols.assign( "save", CallablePythonValue( function_save ) )
        context.symbols.assign( "getcwd", CallablePythonValue( function_getcwd ) )

        context.symbols.assign( "bool", CallablePythonValue( bool ) )
        context.symbols.assign( "int", CallablePythonValue( int ) )
        context.symbols.assign( "float", CallablePythonValue( float ) )
        context.symbols.assign( "str", CallablePythonValue( str ) )
        context.symbols.assign( "list", CallablePythonValue( list ) )
        context.symbols.assign( "dict", CallablePythonValue( dict ) )
        context.symbols.assign( "range", CallablePythonValue( range ) )
        context.symbols.assign( "enumerate", CallablePythonValue( enumerate ) )
        context.symbols.assign( "len", CallablePythonValue( function_len ) )
        context.symbols.assign( "map", CallablePythonValue( map ) )
        context.symbols.assign( "filter", CallablePythonValue( filter ) )

        context.symbols.assign( "mod", CallablePythonValue( function_mod ) )
        context.symbols.assign( "div", CallablePythonValue( function_div ) )

        context.symbols.assign( "inspect_context", CallablePythonValue( function_inspect_context ) )
        context.symbols.assign( "ord", CallablePythonValue( function_ord ) )
        context.symbols.assign( "chr", CallablePythonValue( function_chr ) )
        context.symbols.assign( "setvar", CallablePythonValue( function_setvar ) )
        context.symbols.assign( "hasattr", CallablePythonValue( function_hasattr ) )
        context.symbols.assign( "getattr", CallablePythonValue( function_getattr ) )
        context.symbols.assign( "setattr", CallablePythonValue( function_setattr ) )
        context.symbols.assign( "gettime", CallablePythonValue( function_gettime ) )
        context.symbols.assign( "settime", CallablePythonValue( function_settime ) )
        context.symbols.assign( "cc", CallablePythonValue( function_cc ) )
        context.symbols.assign( "setvoice", CallablePythonValue( function_setvoice ) )
        context.symbols.assign( "setinstrument", CallablePythonValue( function_setinstrument ) )
        context.symbols.assign( "sfload", CallablePythonValue( function_sfload ) )
        context.symbols.assign( "sfunload", CallablePythonValue( function_sfunload ) )
        context.symbols.assign( "interval", Interval )
        context.symbols.assign( "scale", Scale )

        context.symbols.assign( "sequencers\\ABC", CallablePythonValue( seqs.ABCSequencer ) )
        context.symbols.assign( "sequencers\\PDF", CallablePythonValue( seqs.PDFSequencer ) )
        context.symbols.assign( "sequencers\\HTML", CallablePythonValue( seqs.HTMLSequencer ) )
        context.symbols.assign( "sequencers\\Midi", CallablePythonValue( seqs.MidiSequencer ) )
        context.symbols.assign( "sequencers\\Debug", CallablePythonValue( seqs.DebugSequencer ) )
        context.symbols.assign( "sequencers\\FluidSynth", CallablePythonValue( seqs.FluidSynthSequencer ) )

        context.symbols.assign( "Music", CallablePythonValue( Music ) )
        context.symbols.assign( "Metronome", CallablePythonValue( Metronome ) )
        context.symbols.assign( "met", CallablePythonValue( Metronome ) )

        context.symbols.assign( "voices\\create", CallablePythonValue( function_voices_create ) )

        # TODO: Make tests expose a script object
        if context.script is not None:
            context.script.import_module( self.context, Path( __file__ ).parent / "std.mkl", save_cache = False )



from typing import List, Optional, Tuple
from musikla.core import Context, Value
from ..node import Node
from ..expressions import ExpressionNode

# Since returns values of kind OBJECT, this node is not ready for primetime yet.
# Used as a placeholder for the keyboard creation macro
class ListComprehensionNode(ExpressionNode):
    def __init__ ( self, expression : Node, variable : List[str], min : Node, max : Node, condition : Node = None, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.expression : Node = expression
        self.variable : List[str] = variable
        self.min : Node = min
        self.max : Node = max
        self.condition : Optional[Node] = condition
    
    def __eval__ ( self, context, assignment : bool = False ): 
        return None



from musikla.parser.printer import CodePrinter
from typing import Tuple
from ..node import Node
from .expression_node import ExpressionNode
from musikla.core import Context

class UnaryOperatorNode( ExpressionNode ):
    def __init__ ( self, node : Node, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.node : Node = node

    def __eval__ ( self, context, assignment : bool = False ):
        return None

class NotOperatorNode ( UnaryOperatorNode ):
    def __eval__ ( self, context : Context, assignment : bool = False ):
        value = self.node.eval( context )

        return not bool( value )
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( "not " )

        self.node.to_source( printer )



from musikla.parser.printer import CodePrinter
from typing import Tuple
from musikla.core import Context
from .statement_node import StatementNode

class ImportStatementNode( StatementNode ):
    def __init__ ( self, module_name : str, local : bool, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.module_name : str = module_name
        self.local : bool = local

    def __eval__ ( self, context : Context ):
        module_path = context.script.resolve_import( context.symbols.lookup( '__file__' ), self.module_name, local = self.local )
        
        context.script.import_module( context, module_path )

        return None

    def to_source ( self, printer : CodePrinter ):
        printer.add_token( 'import ' )

        if self.local:
            printer.add_token( f'"{self.module_name}"' )
        else:
            printer.add_token( self.module_name )



from musikla.parser.printer import CodePrinter
from .music_node import MusicNode
from musikla.core.theory import Chord, Interval
from musikla.core.events import ChordEvent
from typing import Tuple

class ChordNode( MusicNode ):
    def __init__ ( self, chord : Chord, position : Tuple[int, int, int] = None ):
        super().__init__( position )

        self.chord : Chord = chord
    
    def to_source ( self, printer : CodePrinter ):
        printer.add_token( self.chord.to_string( base_octave = 0 ) )

    def get_events ( self, context ):
        o = Interval.octaves_to_semitones( context.voice.octave )

        chord = ChordEvent(
            timestamp = context.cursor,
            pitches = [ p + o for p in self.chord.to_pitches() ],
            duration = context.get_duration( self.chord.value ),
            value = context.get_value( self.chord.value ),
            voice = context.voice,
            velocity = context.voice.velocity,
            staff = 0
        )

        context.cursor += chord.duration

        yield chord
